/*
 * Attack wrapper
 *
 * This program is used to launch the attack program, to provide it with
 * traces from the public database (or from the private database when run
 * inside Telecom Paristech) and to collect results.
 *
 *
 * Copyright (C) 2009, 2014 Guillaume Duc <guillaume.duc@telecom-paristech.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */


#include <stdint.h>

#include <iostream>
#include <iomanip>
#include <cstdlib>

#include <unistd.h>
#include <getopt.h>

#include "attack_wrapper.h"

// Long options ids
static const int COMPATIBILITY_V2_OPT = 512;
static const int PROVIDE_OFFSET_V4_RSM_OPT = 513;
static const int DO_NOT_CHECK_TRACES_AVAILABILITY_OPT = 514;
static const int PROVIDE_OFFSETS_V4_2_OPT = 515;
static const int NO_FLOAT_FORMAT_V4_2_OPT = 516;

// Contest id string
static const std::string V2_ID = "v2";
static const std::string V4_RSM_ID = "v4_RSM";
static const std::string V4_2_ID = "v4_2";


Attack_Wrapper::Attack_Wrapper ()
{
  cl_num_iterations = 0;
  cl_key_num = 0;

#ifdef _WIN32
  cl_fifo_mode = 1; // FIFO mode is enabled by default on Windows
#else /* _WIN32 */
  cl_fifo_mode = 0; // Fork mode is enabled by default on Unix
#endif /* _WIN32 */

  cl_truncate = 0;

  contest = NULL;

  cl_comm_compatibility_v2 = false;
  cl_provide_offset_v4_rsm = false;
  cl_provide_offsets_v4_2 = false;
  cl_check_traces_availability = true;
  cl_use_float_format = true;
}


/**
 * Parse the command line options
 */
void Attack_Wrapper::parse_command_line (int argc, char **argv)
{
  int c;

  while (1)
    {
      static struct option options[] =
	{
	  {"iterations", required_argument, NULL, 'i'},
	  {"key", required_argument, NULL, 'k'},
	  {"output", required_argument, NULL, 'o'},
	  {"trace_dir", required_argument, NULL, 'd'},
	  {"dir_index", required_argument, NULL, 'x'},
	  {"help", no_argument, NULL, 'h'},
	  {"version", no_argument, NULL, 'v'},
	  {"fifo", no_argument, &cl_fifo_mode, 1},
	  {"truncate", no_argument, &cl_truncate, 1},
	  {"edition", required_argument, NULL, 'e'},
	  {"compatibility_v2", no_argument, NULL, COMPATIBILITY_V2_OPT},
	  {"provide_offset_v4_rsm", no_argument, NULL, PROVIDE_OFFSET_V4_RSM_OPT},
	  {"provide_offsets_v4_2", no_argument, NULL, PROVIDE_OFFSETS_V4_2_OPT},
	  {"do_not_check_traces_availability", no_argument, NULL, DO_NOT_CHECK_TRACES_AVAILABILITY_OPT},
	  {"no_float_format", no_argument, NULL, NO_FLOAT_FORMAT_V4_2_OPT},
	  {NULL, 0, NULL, 0}
	};

      int option_index = 0;

      c = getopt_long (argc, argv, "i:k:o:d:x:hvfte:", options, &option_index);

      if (c == -1)
	break;

      switch (c)
	{
	case 'i':
	  cl_num_iterations = atoi (optarg);
	  break;
	case 'k':
	  cl_key_num = atoi (optarg);
	  break;
	case 'o':
	  cl_output_filename = optarg;
	  break;
	case 'd':
	  cl_trace_dir = optarg;
	  break;
	case 'x':
	  cl_index_filename = optarg;
	  break;
	case 'h':
	  print_usage ();
	  exit (EXIT_SUCCESS);
	  break;
	case 'v':
	  print_version ();
	  exit (EXIT_SUCCESS);
	  break;
	case 'f':
	  cl_fifo_mode = 1;
	  break;
	case 't':
	  cl_truncate = 1;
	  break;
	case 'e':
	  cl_edition = optarg;
	  break;
	case COMPATIBILITY_V2_OPT:
	  cl_comm_compatibility_v2 = true;
	  break;
	case PROVIDE_OFFSET_V4_RSM_OPT:
	  cl_provide_offset_v4_rsm = true;
	  break;
	case PROVIDE_OFFSETS_V4_2_OPT:
	  cl_provide_offsets_v4_2 = true;
	  break;
	case DO_NOT_CHECK_TRACES_AVAILABILITY_OPT:
	  cl_check_traces_availability = false;
	  break;
	case NO_FLOAT_FORMAT_V4_2_OPT:
	  cl_use_float_format = false;
	  break;
	}
    }

  if (optind >= argc)
    {
      std::cerr << "E - Argument missing!" << std::endl;
      print_usage ();
      exit (EXIT_FAILURE);
    }

  cl_attack_program_filename = argv[optind];


  // Do some checks
  if (cl_num_iterations < 0)
    {
      std::cerr << "E - The number of iterations is out of range!" << std::endl;
      exit (EXIT_FAILURE);
    }

  if (cl_trace_dir.empty ())
    {
      std::cerr << "E - No trace directory supplied!" << std::endl;
      std::cerr << "E - Please use the --trace-dir option." << std::endl;
      exit (EXIT_FAILURE);
    }
  if (cl_index_filename.empty ())
    {
      std::cerr << "E - No index file supplied!" << std::endl;
      std::cerr << "E - Please use the --dir-index option." << std::endl;
      exit (EXIT_FAILURE);
    }
  if (cl_edition.empty ())
    {
      std::cerr << "E - No edition supplied!" << std::endl;
      std::cerr << "E - Please use the --edition option." << std::endl;
      exit (EXIT_FAILURE);
    }

  if (cl_output_filename.empty ())
    cl_output_filename = "results";

  // Specific options for v2
  if ((cl_edition != V2_ID) && cl_comm_compatibility_v2)
    {
      std::cerr << "E - The option \"compatibility_v2\" is only available for edition \"" << V2_ID << "\"!" << std::endl;
      exit (EXIT_FAILURE);
    }

  // Specific options for v4_RSM
  if ((cl_edition != V4_RSM_ID) && cl_provide_offset_v4_rsm)
    {
      std::cerr << "E - The option \"provide_offset_v4_rsm\" is only available for edition \""<< V4_RSM_ID << "\"!" << std::endl;
      exit (EXIT_FAILURE);
    }

  // Specific options for v4_2
  if ((cl_edition != V4_2_ID) && cl_provide_offsets_v4_2)
    {
      std::cerr << "E - The option \"provide_offsets_v4_2\" is only available for edition \""<< V4_2_ID << "\"!" << std::endl;
      exit (EXIT_FAILURE);
    }
  if ((cl_edition != V4_2_ID) && !cl_use_float_format)
    {
      std::cerr << "E - The option \"no_float_format\" is only available for edition \""<< V4_2_ID << "\"!" << std::endl;
      exit (EXIT_FAILURE);
    }

  // Outputs the options chosen
  std::cout << "D - Output filename = " << cl_output_filename;
  std::cout << (cl_truncate ? " (truncate if exists)" : " (abort if exists)") << std::endl;

  if (cl_fifo_mode == 0)
    {
      std::cout << "D - Normal mode (fork mode)" << std::endl;
      std::cout << "D - Attack program = " << cl_attack_program_filename << std::endl;
    }
  else
    {
      std::cout << "D - FIFO mode" << std::endl;
      std::cout << "D - Base name for FIFOs = " << cl_attack_program_filename << std::endl;
    }

  std::cout << "D - Compatibility mode (v2) = " << (cl_comm_compatibility_v2 ? "enabled" : "disabled") << std::endl;

  std::cout << "D - Traces will be read from directory " << cl_trace_dir << std::endl;
  std::cout << "D - Using index file " << cl_index_filename << std::endl;

  if (cl_check_traces_availability)
    std::cout << "D - We will check if traces are available" << std::endl;
  else
    std::cout << "D - We assume that all the traces in the index file are available" << std::endl;

  if (cl_edition == V4_RSM_ID)
    {
      if (cl_provide_offset_v4_rsm)
	std::cout << "D - Offsets are provided to the attack" << std::endl;
      else
	std::cout << "D - Offsets are not provided to the attack" << std::endl;
    }

  if (cl_edition == V4_2_ID)
    {
      if (cl_provide_offsets_v4_2)
	std::cout << "D - Offsets/Shuffles are provided to the attack" << std::endl;
      else
	std::cout << "D - Offsets/Shuffles are not provided to the attack" << std::endl;

      if (cl_use_float_format)
	std::cout << "D - Samples are transfered as floats" << std::endl;
      else
	std::cout << "D - Samples are transfered as integers" << std::endl;
    }
}


/**
 * Print the command line options
 */
void Attack_Wrapper::print_usage () const
{
  std::cout << "Usage: attack_wrapper [OPTION]... FILENAME" << std::endl << std::endl;

  std::cout << "Launch the attack program." << std::endl;
  std::cout << "In fork mode, FILENAME represents the name of the executable" << std::endl;
  std::cout << "file containing the attack program." << std::endl;
  std::cout << "In FIFO mode, FILENAME represents the base name of the FIFOs" << std::endl << std::endl;

  std::cout << "The options are:" << std::endl;
  std::cout << "-i, --iterations=NUM\t\tProvide NUM traces to the attack program (default: all available traces)" << std::endl;
  std::cout << "-k, --key=NUM\t\t\tUse the key number NUM in the base (default: first key)" << std::endl;
  std::cout << "-o, --output=FILENAME\t\tStore the results of the attack in FILENAME (default: results)" << std::endl;
  std::cout << "-t, --truncate\t\t\tAllow to overwrite an existing results file (default: no)" << std::endl;
  std::cout << "-d, --trace-dir=DIR\t\tPath to the directory containg the traces" << std::endl;
  std::cout << "-x, --dir-index=FILENAME\tPath to the trace directory index file" << std::endl;
  std::cout << "-f, --fifo\t\t\tFIFO mode (FILENAME represents the base name for the 2 FIFOs)" << std::endl;
  std::cout << "-e, --edition=EDITION\t\tSelect the version of the contest (see below)" << std::endl;
  std::cout << "-h, --help\t\t\tPrint this help message" << std::endl;
  std::cout << "-v, --version\t\t\tPrint the version of the attack wrapper" << std::endl;
  std::cout << "--do_not_check_traces_availability\tDo not check if traces listed in the index file are available" << std::endl << std::endl;

  std::cout << "List of available EDITIONs:" << std::endl;
  std::cout << "\"" << V2_ID << "\"\t\t - DPA contest v2 (AES-128 on FPGA)" << std::endl;
  std::cout << "\"" << V4_RSM_ID << "\"\t - DPA contest v4, AES256 RSM on smartcard " << std::endl;
  std::cout << "\"" << V4_2_ID << "\"\t\t - DPA contest v4.2, AES128 RSM Improved on smartcard " << std::endl << std::endl;

  std::cout << "Specific options to " << V2_ID << ":" << std::endl;
  std::cout << "--compatibility_v2\tUsed to run attacks designed for the first version of the wrapper" << std::endl << std::endl;

  std::cout << "Specific options to " << V4_RSM_ID << ":" << std::endl;
  std::cout << "--provide_offset_v4_rsm\tProvide the correct offset to the attack program" << std::endl << std::endl;

  std::cout << "Specific options to " << V4_2_ID << ":" << std::endl;
  std::cout << "--provide_offsets_v4_2\tProvide the correct offsets, shuffle0 and shuffle10 to the attack program" << std::endl;
  std::cout << "--no_float_format\tTransfer samples as integers to the attack (old behavior) instead of floats" << std::endl << std::endl;
}


/*
 * This function displays the version of the attack wrapper.
 */

void Attack_Wrapper::print_version () const
{
  std::cout << "Attack wrapper" << std::endl;
  std::cout << "Version: " << PACKAGE_VERSION << std::endl;
}


/**
 * Run the contest
 */
void Attack_Wrapper::run_contest ()
{
  bool res;

  if (cl_edition == V2_ID)
    contest = new Contest_v2 (cl_trace_dir,
			      cl_index_filename,
			      cl_output_filename,
			      cl_truncate,
			      (! cl_fifo_mode),
			      cl_attack_program_filename,
			      cl_comm_compatibility_v2,
			      cl_check_traces_availability);

  if (cl_edition == V4_RSM_ID)
    contest = new Contest_v4_RSM (cl_trace_dir,
				  cl_index_filename,
				  cl_output_filename,
				  cl_truncate,
				  (! cl_fifo_mode),
				  cl_attack_program_filename,
				  cl_provide_offset_v4_rsm,
				  cl_check_traces_availability);

  if (cl_edition == V4_2_ID)
    contest = new Contest_v4_2 (cl_trace_dir,
				cl_index_filename,
				cl_output_filename,
				cl_truncate,
				(! cl_fifo_mode),
				cl_attack_program_filename,
				cl_provide_offsets_v4_2,
				cl_check_traces_availability,
				cl_use_float_format);

  if (contest == NULL)
    {
      std::cerr << "E - Unknown version of the contest!" << std::endl;
      return;
    }

  std::cout << "D - Key # = " << std::dec << cl_key_num << std::endl;

  res = contest->set_key_num (cl_key_num);

  if (! res)
    return;

  cl_num_iterations = contest->set_num_iterations (cl_num_iterations);

  std::cout << "D - # of traces = " << std::dec << cl_num_iterations << std::endl;

  contest->run_contest ();
}


/*
 * The main function of the program.
 */
int
main (int argc, char **argv)
{
  Attack_Wrapper wrapper;

  wrapper.parse_command_line (argc, argv);

  wrapper.run_contest ();

  exit (EXIT_SUCCESS);
}
