/*
 * Classes for communication with the attack program
 *
 * Copyright (C) 2013 Guillaume Duc <guillaume.duc@telecom-paristech.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <iostream>
#include <iomanip>
#include <cstring>
#include <cerrno>
#include <cassert>
#include <cstdlib>
#include <limits>

#include <unistd.h>

#ifndef _WIN32
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <signal.h>
#endif /* _WIN32 */

#include "comm.h"



/*******************************************************************
 * Class Attack_Comm
 *******************************************************************/


/**
 * Constructor
 *
 * \arg \c trace_length Length (in bytes) of the raw trace data
 * \arg \c results_length Length (in bytes) of the results sent by the attack program
 */
Attack_Comm::Attack_Comm (unsigned int trace_length,
			  unsigned int results_length)
  : trace_length (trace_length),
    results_length (results_length)
{
  last_trace = false;
  last_results = false;
}


/**
 * Send a trace to the attack
 *
 * \arg \c data Pointer to the raw representation of the trace
 * \arg \c last_trace Is it the last trace to send
 * \return \c true if the trace has been properly sent
 */
bool Attack_Comm::send_trace (const int8_t *data, bool last_trace)
{
  this->last_trace = last_trace;
  return send_data (data, trace_length);
}


/**
 * Receive a result from the attack
 *
 * \arc \c results Pointer to a buffer where the raw results will be written
 * \arg \c last_results Is it the last results expected
 * \return \c true if the result has been properly received
 */
bool Attack_Comm::receive_results (int8_t *results, bool last_results)
{
  this->last_results = last_results;
  return receive_data (results, results_length);
}



/*******************************************************************
 * Class Attack_Comm_FIFO
 *******************************************************************/


/**
 * Constructor
 *
 * \arg \c trace_length Length (in bytes) of the raw trace data
 * \arg \c results_length Length (in bytes) of the results sent by the attack program
 * \arg \c fifo_base_name Prefix used to create the filenames of the FIFOs
 */
Attack_Comm_FIFO::Attack_Comm_FIFO (unsigned int trace_length,
				    unsigned int results_length,
				    std::string fifo_base_name,
				    bool compatibility_v2)
  : Attack_Comm (trace_length, results_length)
{
#ifdef _WIN32
  fifo_output_filename = "\\\\.\\pipe\\" + fifo_base_name + "_from_wrapper";
  fifo_input_filename = "\\\\.\\pipe\\" + fifo_base_name + "_to_wrapper";
#else /* _WIN32 */
  fifo_output_filename = fifo_base_name + "_from_wrapper";
  fifo_input_filename = fifo_base_name + "_to_wrapper";
#endif /* _WIN32 */

  output_fd = -1;
  input_fd = -1;
  this->compatibility_v2 = compatibility_v2;
}


#ifdef _WIN32

/**
 * Send data to the attack (Windows version)
 *
 * \arg \c data Pointer to the data to send
 * \arg \c length Length (in bytes) of the data to send
 * \return \c true if the data has been properly sent
 */
bool Attack_Comm_FIFO::send_data (const int8_t *data, unsigned int length)
{
  bool res;
  DWORD bytes_written;

  res = WriteFile (output_pipe_handle,
		   data, length, &bytes_written, NULL);

  if (! res)
    {
      std::cerr << "E - Error writting to FIFO!" << std::endl;
      std::cerr << "E - Reason: " << GetLastError () << std::endl;
      return false;
    }
  if (bytes_written < length)
    {
      std::cerr << "E - Error writting to FIFO!" << std::endl;
      std::cerr << "Only " << bytes_written << " bytes written instead of " << length << std::endl;
      return false;
    }

  return true;
}

#else /* _WIN32 */

/**
 * Send data to the attack (Unix version)
 *
 * \arg \c data Pointer to the data to send
 * \arg \c length Length (in bytes) of the data to send
 * \return \c true if the data has been properly sent
 */
bool Attack_Comm_FIFO::send_data (const int8_t *data, unsigned int length)
{
  size_t total_bytes_written = 0;
  ssize_t bytes_written;

  assert (output_fd >= 0);

  while (total_bytes_written < length)
    {
      bytes_written = write (output_fd,
			     ((const char *) data) + total_bytes_written,
			     length - total_bytes_written);

      if (bytes_written == -1)
	{
	  if (errno == EAGAIN)
	    continue;

	  std::cerr << "E - Error while sending data to the attack (send_data)" << std::endl;
	  std::cerr << "E - Reason: " << std::strerror (errno) << std::endl;

	  return false;
	}

      total_bytes_written += bytes_written;
    }

  return true;
}

#endif /* _WIN32 */



#ifdef _WIN32

/**
 * Receive data from the attack (Windows version)
 *
 * \arg \c data Pointer to a buffer where the data will be written
 * \arg \c length Length (in bytes) of the data to receive
 * \return \c true if the data has been properly received
 */
bool Attack_Comm_FIFO::receive_data (int8_t *data, unsigned int length)
{
  bool res;
  DWORD bytes_read;
  unsigned int total_bytes_read = 0;

  while (total_bytes_read < length)
    {
      res = ReadFile (input_pipe_handle,
		      data + total_bytes_read,
		      length - total_bytes_read,
		      &bytes_read,
		      NULL);

      if (! res)
	{
	  std::cerr << "E - Error reading from FIFO!" << std::endl;
	  std::cerr << "E - Reason: " << GetLastError () << std::endl;
	  return false;
	}

      total_bytes_read += bytes_read;
    }

  return true;
}

#else /* _WIN32 */

/**
 * Receive data from the attack (Unix version)
 *
 * \arg \c data Pointer to a buffer where the data will be written
 * \arg \c length Length (in bytes) of the data to receive
 * \return \c true if the data has been properly received
 */
bool Attack_Comm_FIFO::receive_data (int8_t *data, unsigned int length)
{
  size_t total_bytes_read = 0;
  ssize_t bytes_read;

  assert (input_fd >= 0);

  while (total_bytes_read < length)
    {
      bytes_read = read (input_fd,
			 ((char *) data) + total_bytes_read,
			 length - total_bytes_read);

      if (bytes_read == -1)
	{
	  if (errno == EAGAIN)
	    continue;

	  std::cerr << "E - Error while reading data from the attack (receive_data)" << std::endl;
	  std::cerr << "E - Reason: " << std::strerror (errno) << std::endl;

	  return false;
	}
      total_bytes_read += bytes_read;
    }

  return true;
}

#endif /* _WIN32 */



#ifdef _WIN32
/**
 * Prepare the communications with the attack program. In this case,
 * create the FIFO (Windows version)
 *
 * \return \c true if everything is fine
 */
bool Attack_Comm_FIFO::prepare_comm ()
{
  // Create the output FIFO
  output_pipe_handle = CreateNamedPipe (fifo_output_filename.c_str (),
					PIPE_ACCESS_OUTBOUND,
					PIPE_TYPE_BYTE,
					1,
					0,
					0,
					0,
					NULL);

  if ((output_pipe_handle == NULL) || (output_pipe_handle == INVALID_HANDLE_VALUE))
    {
      std::cerr << "E - Error during creation of FIFO " << fifo_output_filename << std::endl;
      std::cerr << "E - Reason: " << GetLastError () << std::endl;

      return false;
    }

  // Create the input FIFO
  input_pipe_handle = CreateNamedPipe (fifo_input_filename.c_str (),
				       PIPE_ACCESS_INBOUND,
				       PIPE_TYPE_BYTE,
				       1,
				       0,
				       0,
				       0,
				       NULL);

  if ((input_pipe_handle == NULL) || (input_pipe_handle == INVALID_HANDLE_VALUE))
    {
      std::cerr << "E - Error during creation of FIFO " << fifo_input_filename << std::endl;
      std::cerr << "E - Reason: " << GetLastError () << std::endl;

      return false;
    }

  std::cout << "I - FIFO Wrapper -> Attack: " << fifo_output_filename << std::endl;
  std::cout << "I - FIFO Attack -> Wrapper: " << fifo_input_filename << std::endl;

  return true;
}

#else /* _WIN32 */

/**
 * Prepare the communications with the attack program. In this case,
 * create the FIFO (Unix version)
 *
 * \return \c true if everything is fine
 */
bool Attack_Comm_FIFO::prepare_comm ()
{
  // Create the two FIFOs
  if (mkfifo (fifo_output_filename.c_str (), 0600) < 0)
    {
      std::cerr << "E - Error during creation of FIFO " << fifo_output_filename << std::endl;
      std::cerr << "E - Reason: " << std::strerror (errno) << std::endl;

      return false;
    }

  if (mkfifo (fifo_input_filename.c_str (), 0600) < 0)
    {
      std::cerr << "E - Error during creation of FIFO " << fifo_input_filename << std::endl;
      std::cerr << "E - Reason: " << std::strerror (errno) << std::endl;

      return false;
    }

  std::cout << "I - FIFO Wrapper -> Attack: " << fifo_output_filename << std::endl;
  std::cout << "I - FIFO Attack -> Wrapper: " << fifo_input_filename << std::endl;


  // The output fifo (wrapper->attack) should be non blocking
  // so that writes to it won't block
  output_fd = open (fifo_output_filename.c_str (), O_RDWR | O_NONBLOCK);
  if (output_fd < 0)
    {
      std::cerr << "E - Error while opening the FIFO " << fifo_output_filename << std::endl;
      std::cerr << "E - Reason: " << std::strerror (errno) << std::endl;

      return false;
    }


  // The input fifo (attack->wrapper) should be blocking
  // so that reads will block until data is available,
  // but we open it as O_RDWR to prevent open from blocking
  input_fd = open (fifo_input_filename.c_str (), O_RDWR);
  if (input_fd < 0)
    {
      std::cerr << "E - Error while opening the FIFO " << fifo_input_filename << std::endl;
      std::cerr << "E - Reason: " << std::strerror (errno) << std::endl;

      return false;
    }


  // Ignore SIGPIPE (we will detect unexpected attack termination
  // by analysing the return of write (EPIPE)
  struct sigaction sa;
  sa.sa_handler = SIG_IGN;
  sa.sa_flags = 0;
  sigemptyset (&sa.sa_mask);
  sigaction (SIGPIPE, &sa, NULL);

  return true;
}
#endif /* _WIN32 */


/**
 * Launch the attack and wait until the attack is ready
 *
 * \arg \c num_traces Number of traces that will be sent to the attack
 * \return \c true if the attack has been properly launched
 */
bool Attack_Comm_FIFO::launch_attack_and_wait (uint32_t num_traces)
{
  bool res;
  int8_t data;
  int state = 0;

#ifdef _WIN32
  // Wait for connection to the pipes
  // TODO
  // Be careful here, the client shall open the FIFOs in the same order... ?

  res = ConnectNamedPipe (output_pipe_handle, NULL);
  if (! res)
    {
      DWORD err = GetLastError ();

      if (err == ERROR_PIPE_CONNECTED)
	{
	  std::cout << "I - Client already connected to the FIFO " << fifo_output_filename << " (ok)" << std::endl;
	}
      else
	{
	  std::cerr << "E - Error during during the connection to the FIFO " << fifo_output_filename << std::endl;
	  std::cerr << "E - Reason: " << err << std::endl;
	  return false;
	}
    }

  res = ConnectNamedPipe (input_pipe_handle, NULL);
  if (! res)
    {
      DWORD err = GetLastError ();

      if (err == ERROR_PIPE_CONNECTED)
	{
	  std::cout << "I - Client already connected to the FIFO " << fifo_input_filename << " (ok)" << std::endl;
	}
      else
	{
	  std::cerr << "E - Error during during the connection to the FIFO " << fifo_input_filename << std::endl;
	  std::cerr << "E - Reason: " << err << std::endl;
	  return false;
	}
    }

#else  /* _WIN32 */
  assert (input_fd >= 0);
  assert (output_fd >= 0);
#endif /* _WIN32 */


  // Send the number of iterations (i.e. the number of traces)
  std::cout << "I - Sending # of iterations (" << num_traces << ")" << std::endl;

  if (compatibility_v2)
    {
      uint16_t tr = num_traces;
      assert (num_traces < std::numeric_limits<uint16_t>::max ());
      res = send_data ((int8_t *) &tr, sizeof (tr)); // uint16_t
    }
  else
    res = send_data ((int8_t *) &num_traces, sizeof (num_traces)); // uint32_t

  if (! res)
    return false;

  // Wait for the attack (string LF.LF (hex: 0A 2E 0A))
  while (true)
    {
      res = receive_data (&data, sizeof (data));
      if (! res)
	return false;

      switch (state)
	{
	case 0:
	  if (data == 0x0A)
	    state = 1;
	  break;
	case 1:
	  if (data == 0x2E)
	    state = 2;
	  else
	    state = 0;
	  break;
	case 2:
	  if (data == 0x0A)
	    state = 3;
	  else
	    state = 0;
	  break;
	}

      if (state == 3)
	break;
    }

  std::cout << "I - The attack is ready!" << std::endl;

  return true;
}


/**
 * Close the FIFOs
 *
 * \return \c true if ok
 */
bool Attack_Comm_FIFO::close_comm ()
{
#ifdef _WIN32

  CloseHandle (input_pipe_handle);
  CloseHandle (output_pipe_handle);

  std::cout << "I - FIFOs closed" << std::endl;

#else /* _WIN32 */

  // Close the two FIFOs
  if (output_fd >= 0)
    {
      close (output_fd);
      unlink (fifo_output_filename.c_str ());
    }
  if (input_fd >= 0)
    {
      close (input_fd);
      unlink (fifo_input_filename.c_str ());
    }

  output_fd = -1;
  input_fd = -1;

  std::cout << "I - FIFOs closed and removed" << std::endl;

#endif /* _WIN32 */

  return true;
}



#ifndef _WIN32

/*******************************************************************
 * Class Attack_Comm_Fork
 *******************************************************************/

static Attack_Comm_Fork * main_attack_comm_fork = NULL;


/**
 * Constructor
 *
 * \arg \c trace_length Length (in bytes) of the raw trace data
 * \arg \c results_length Length (in bytes) of the results sent by the attack program
 * \arg \c attack_program_full_name Name and path of the attack program to launch
 */
Attack_Comm_Fork::Attack_Comm_Fork (unsigned int trace_length,
				    unsigned int results_length,
				    std::string attack_program_full_name,
				    bool compatibility_v2)
  : Attack_Comm (trace_length, results_length),
    attack_program_full_name (attack_program_full_name)
{
  output_fd = -1;
  input_fd = -1;

  stdin_pipe_fd[0] = -1;
  stdin_pipe_fd[1] = -1;

  stdout_pipe_fd[0] = -1;
  stdout_pipe_fd[1] = -1;

  attack_dead = false;

  this->compatibility_v2 = compatibility_v2;

  // Extract program filename from full name
  unsigned int pos = attack_program_full_name.find_last_of ("/");
  attack_program_file_name = attack_program_full_name.substr (pos + 1);
}


/**
 * Send data to the attack
 *
 * \arg \c data Pointer to the data to send
 * \arg \c length Length (in bytes) of the data to send
 * \return \c true if the data has been properly sent
 */
bool Attack_Comm_Fork::send_data (const int8_t *data, unsigned int length)
{
  size_t total_bytes_written = 0;
  ssize_t bytes_written;

  assert (output_fd >= 0);

  if (attack_dead)
    return false;

  while (total_bytes_written < length)
    {
      bytes_written = write (output_fd,
			     ((const char *) data) + total_bytes_written,
			     length - total_bytes_written);

      if (bytes_written == -1)
	{
	  if (errno == EAGAIN)
	    continue;

	  std::cerr << "E - Error while sending data to the attack (send_data)" << std::endl;
	  std::cerr << "E - Reason: " << std::strerror (errno) << std::endl;

	  return false;
	}

      total_bytes_written += bytes_written;
    }

  return true;
}


/**
 * Receive data from the attack
 *
 * \arg \c data Pointer to a buffer where the data will be written
 * \arg \c length Length (in bytes) of the data to receive
 * \return \c true if the data has been properly received
 */
bool Attack_Comm_Fork::receive_data (int8_t *data, unsigned int length)
{
  size_t total_bytes_read = 0;
  ssize_t bytes_read;

  assert (input_fd >= 0);

  if (attack_dead)
    return false;

  while (total_bytes_read < length)
    {
      bytes_read = read (input_fd,
			 ((char *) data) + total_bytes_read,
			 length - total_bytes_read);

      if (bytes_read == -1)
	{
	  if (errno == EAGAIN)
	    continue;

	  std::cerr << "E - Error while reading data from the attack (receive_data)" << std::endl;
	  std::cerr << "E - Reason: " << std::strerror (errno) << std::endl;

	  return false;
	}
      total_bytes_read += bytes_read;
    }

  return true;
}


/**
 * Prepare the communications with the attack program.
 *
 * \return \c true if everything is fine
 */
bool Attack_Comm_Fork::prepare_comm ()
{
  // Create two PIPE for communication with the attack program
  if (pipe (stdin_pipe_fd) == -1)
    {
      std::cerr << "E - Error while creating pipes for communication with the attack" << std::endl;
      std::cerr << "E - Reason: " << std::strerror (errno) << std::endl;

      return false;
    }
  if (pipe (stdout_pipe_fd) == -1)
    {
      std::cerr << "E - Error while creating pipes for communication with the attack" << std::endl;
      std::cerr << "E - Reason: " << std::strerror (errno) << std::endl;

      return false;
    }

  output_fd = stdin_pipe_fd[1];
  input_fd = stdout_pipe_fd[0];

  // Set an handler for SIGCHLD (termination of the child, i.e. the attack program)
  main_attack_comm_fork = this;

  struct sigaction sa;
  sa.sa_handler = attack_comm_fork_sig_chld_handler;
  sa.sa_flags = 0;
  sigemptyset (&sa.sa_mask);

  sigaction (SIGCHLD, &sa, NULL);

  return true;
}


/**
 * Launch the attack and wait until the attack is ready
 *
 * \arg \c num_traces Number of traces that will be sent to the attack
 * \return \c true if the attack has been properly launched
 */
bool Attack_Comm_Fork::launch_attack_and_wait (uint32_t num_traces)
{
  bool res;
  int8_t data;
  int state = 0;
  pid_t process_id;

  assert (input_fd >= 0);
  assert (output_fd >= 0);

  // Fork
  process_id = fork ();

  if (process_id == -1)
    {
      std::cerr << "E - Error while forking to launch the attack program" << std::endl;
      std::cerr << "E - Reason: " << std::strerror (errno) << std::endl;

      return false;
    }
  if (process_id == 0)
    {
      // Child

      // Preparation of file descriptors
      close (stdin_pipe_fd[1]);
      close (stdout_pipe_fd[0]);

      close (0); // stdin
      close (1); // stdout
      close (2); // stderr

      dup2 (stdin_pipe_fd[0], 0);
      dup2 (stdout_pipe_fd[1], 1);

      // Launch of the attack program
      execl (attack_program_full_name.c_str (), attack_program_file_name.c_str (), (char *) NULL);

      // Normally we only reach this part if execl fails
      std::cerr << "E - Error while launching the attack program (execl)" << std::endl;
      std::cerr << "E - Reason: " << std::strerror (errno) << std::endl;

      // KILL
      std::exit (EXIT_FAILURE);
    }

  // Parent

  // Close unused file descriptors
  close (stdin_pipe_fd[0]);
  close (stdout_pipe_fd[1]);


  // Send the number of iterations (i.e. the number of traces)
  std::cout << "I - Sending # of iterations (" << num_traces << ")" << std::endl;

  if (compatibility_v2)
    {
      uint16_t tr = num_traces;
      assert (num_traces < std::numeric_limits<uint16_t>::max ());
      res = send_data ((int8_t *) &tr, sizeof (tr)); // uint16_t
    }
  else
    res = send_data ((int8_t *) &num_traces, sizeof (num_traces)); // uint32_t

  if (! res)
    return false;

  // Wait for the attack (string LF.LF (hex: 0A 2E 0A))
  std::cout << "I - Wainting for the start of the attack" << std::endl;

  while (true)
    {
      res = receive_data (&data, sizeof (data));
      if (! res)
	return false;

      switch (state)
	{
	case 0:
	  if (data == 0x0A)
	    state = 1;
	  break;
	case 1:
	  if (data == 0x2E)
	    state = 2;
	  else
	    state = 0;
	  break;
	case 2:
	  if (data == 0x0A)
	    state = 3;
	  else
	    state = 0;
	  break;
	}

      if (state == 3)
	break;
    }

  std::cout << "I - The attack is ready!" << std::endl;

  return true;
}


/**
 * Close
 *
 * \return \c true if ok
 */
bool Attack_Comm_Fork::close_comm ()
{
  if (output_fd >= 0)
    close (output_fd);

  if (input_fd >= 0)
    close (input_fd);

  output_fd = -1;
  input_fd = -1;

  return true;
}



void attack_comm_fork_sig_chld_handler (int)
{
  if ((main_attack_comm_fork->last_results == false) && (main_attack_comm_fork->last_trace == false))
    {
      std::cerr << "E - Attack program terminates unexpectedly! (SIGCHLD)" << std::endl;
    }

  main_attack_comm_fork->attack_dead = true;
  wait (NULL);
}

#endif /* _WIN32 */
