/*
 * Classes for communication with the attack program
 *
 * Copyright (C) 2013 Guillaume Duc <guillaume.duc@telecom-paristech.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef COMM_H
#define COMM_H

#include <stdint.h>
#include <string>

#ifdef _WIN32
#include <windows.h>
#endif /* _WIN32 */

class Attack_Comm
{
protected:
  unsigned int trace_length;
  unsigned int results_length;

  bool last_trace, last_results;

  Attack_Comm (unsigned int trace_length,
	       unsigned int results_length);

  virtual bool send_data (const int8_t *data, unsigned int length) = 0;
  virtual bool receive_data (int8_t *data, unsigned int length) = 0;

public:
  virtual bool prepare_comm () = 0;
  virtual bool launch_attack_and_wait (uint32_t num_traces) = 0;
  virtual bool send_trace (const int8_t *data, bool last_trace = false);
  virtual bool receive_results (int8_t *results, bool last_results = false);
  virtual bool close_comm () = 0;
};


class Attack_Comm_FIFO : public Attack_Comm
{
  std::string fifo_output_filename;
  std::string fifo_input_filename;

  int output_fd, input_fd;

  bool compatibility_v2;

#ifdef _WIN32
  HANDLE input_pipe_handle;
  HANDLE output_pipe_handle;
#endif /* WIN32 */

  virtual bool send_data (const int8_t *data, unsigned int length);
  virtual bool receive_data (int8_t *data, unsigned int length);

public:
  Attack_Comm_FIFO (unsigned int trace_length,
		    unsigned int results_length,
		    std::string fifo_base_name,
		    bool compatibility_v2 = false);

  virtual bool prepare_comm ();
  virtual bool launch_attack_and_wait (uint32_t num_traces);
  virtual bool close_comm ();
};


#ifndef _WIN32

class Attack_Comm_Fork : public Attack_Comm
{
  std::string attack_program_full_name;
  std::string attack_program_file_name;

  int output_fd, input_fd;

  int stdin_pipe_fd[2];
  int stdout_pipe_fd[2];

  bool compatibility_v2;

  volatile bool attack_dead;

  virtual bool send_data (const int8_t *data, unsigned int length);
  virtual bool receive_data (int8_t *data, unsigned int length);

public:
  Attack_Comm_Fork (unsigned int trace_length,
		    unsigned int results_length,
		    std::string attack_program_full_name,
		    bool compatibility_v2 = false);

  virtual bool prepare_comm ();
  virtual bool launch_attack_and_wait (uint32_t num_traces);
  virtual bool close_comm ();

  friend void attack_comm_fork_sig_chld_handler (int sig);
};

void attack_comm_fork_sig_chld_handler (int);

#endif /* _WIN32 */

#endif /* COMM_H */
