/*
 * Classes for management of the different editions of the contest
 *
 * Copyright (C) 2013 Guillaume Duc <guillaume.duc@telecom-paristech.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <iostream>
#include <iomanip>
#include <vector>
#include <cassert>

#include "editions.h"

#include "utils.h"


/******************************************************************
 * Class Contest
 ******************************************************************/

/**
 * Constructor for the abstract class Contest
 *
 * \arg \c contest_identifier Numeric identifier for the contest
 * \arg \c trace_directory Name of the directory contining the traces
 * \arg \c results_file Name of the file that will contain the results
 * \arg \c allow_overwrite_results_file Do not stop if the results file already exists
 */
Contest::Contest (uint32_t contest_identifier,
		  const std::string &trace_directory,
		  const std::string &results_file,
		  bool allow_overwrite_results_file)
  : contest_identifier (contest_identifier),
    trace_directory (trace_directory),
    results_file (results_file),
    allow_overwrite_results_file (allow_overwrite_results_file)
{
  trace_manager = NULL;
  results_manager = NULL;
  attack_comm = NULL;

  trace_raw_length = 0;
  results_raw_length = 0;

  num_iterations = -1;
  key_num = -1;
}



/******************************************************************
 * Class Contest_v2
 ******************************************************************/

/** Numeric identifier of the implementation of DPA contest v2 */
const uint32_t Contest_v2::contest_v2_id = 1;


/**
 * Constructor
 *
 * \arg \c trace_directory Name of the directory contining the traces
 * \arg \c index_file Name of the index file
 * \arg \c results_file Name of the file that will contain the results
 * \arg \c allow_overwrite_results_file Do not stop if the results file already exists
 * \arg \c fork_mode Communication mode with the attack program (\c true: Fork mode, \c false: FIFO mode)
 * \arg \c attack_program_name If in fork mode, name of the attack program, if in FIFO mode, prefix for the FIFO names
 * \arg \c comm_compatibility_v2 If \c true, send the number of traces as a uint16_t (instead of uint32_t) to keep the compatibility with the old communication protocol. Default is \c false.
 * \arg \c check_traces_availability For each trace in the index file, check whether the trace file exists. Default is \c true.
 */
Contest_v2::Contest_v2 (const std::string &trace_directory,
			const std::string &index_file,
			const std::string &results_file,
			bool allow_overwrite_results_file,
			bool fork_mode,
			const std::string &attack_program_name,
			bool comm_compatibility_v2,
			bool check_traces_availability)
: Contest (contest_v2_id, trace_directory, 
	   results_file, allow_overwrite_results_file)
{
  this->comm_compatibility_v2 = comm_compatibility_v2;

  trace_manager = new Trace_Manager_v2 (trace_directory, index_file,
					check_traces_availability);

  results_manager = new Results_Manager_AES (contest_identifier, 128);

  trace_raw_length = trace_manager->get_trace_raw_length ();
  results_raw_length = results_manager->get_results_raw_length ();

#ifdef _WIN32
  // FIFO mode only for Windows
  attack_comm = new Attack_Comm_FIFO (trace_raw_length, results_raw_length,
				      attack_program_name, comm_compatibility_v2);
#else /* _WIN32 */
  if (fork_mode)
    {
      // Fork mode
      attack_comm = new Attack_Comm_Fork (trace_raw_length, results_raw_length,
					  attack_program_name, comm_compatibility_v2);
    }
  else
    {
      // FIFO mode
      attack_comm = new Attack_Comm_FIFO (trace_raw_length, results_raw_length,
					  attack_program_name, comm_compatibility_v2);
    }
#endif /* _WIN32 */

  trace_buffer = new int8_t[trace_raw_length];
  results_buffer = new int8_t[results_raw_length];
}


/**
 * Run the contest
 *
 * \return \c false if an error happened during the execution
 */
bool Contest_v2::run_contest ()
{
  bool res;

  Trace_Manager_v2 * tm = static_cast<Trace_Manager_v2 *> (trace_manager);

  const Trace_v2 * trace;

  Timer timer;
  uint32_t duration_ms;
  float duration_s;

  assert (key_num >= 0);
  assert (num_iterations > 0);


  // Prepare the results file
  std::cout << "I - Preparing the results file..." << std::endl;
  res = results_manager->prepare_results_file (results_file, allow_overwrite_results_file);
  if (! res)
    return false;

  // Set up the number of traces
  results_manager->set_number_of_traces (num_iterations);

  // Prepare the communication with the program
  res = attack_comm->prepare_comm ();
  if (! res)
    return false;

  // Launch the attack!
  attack_comm->launch_attack_and_wait (num_iterations);
  if (! res)
    return false;

  // Main loop
  for (int trace_num = 0; trace_num < num_iterations; trace_num++)
    {
      std::cout << "I - Trace #" << std::dec << std::setw (6) << std::setfill ('0')
		<< trace_num << ": Reading trace" << std::flush;

      trace = tm->get_next_trace ();

      if (trace == NULL)
	{
	  std::cerr << std::endl << "E - Oops, no more trace to read?!" << std::endl;
	  break;
	}

      std::cout << " / Sending trace" << std::flush;

      trace->get_raw (trace_buffer);

      timer.start ();

      attack_comm->send_trace (trace_buffer,
			       (trace_num == num_iterations - 1));

      std::cout << " / Waiting for results" << std::flush;

      attack_comm->receive_results (results_buffer,
				    (trace_num == num_iterations - 1));

      duration_ms = timer.stop ();

      std::cout << " / Saving results" << std::flush;

      results_manager->save_raw_results (results_buffer, *trace, duration_ms);

      duration_s = (float) duration_ms / 1000.;

      std::cout << " / Done ("
		<< std::fixed << std::setprecision (1) << duration_s
		<< " s) "
		<< results_manager->get_progression_string ()
		<< std::endl;
    }


  // Close the results file
  std::cout << "I - Closing the results file..." << std::endl;
  results_manager->close_results_file ();

  // Close the comm with the attack program
  attack_comm->close_comm ();

  return true;
}


/**
 * Set the number of traces to use for the execution of the attack.
 * Must be called after set_key_num and before run_contest.
 *
 * \arg \c num_iterations Number of traces to use (0 to use all the available traces)
 * \return Number of traces that will really be used (num_iterations except if there are less traces available)
 */
int Contest_v2::set_num_iterations (unsigned int num_iterations)
{
  assert (key_num >= 0);

  Trace_Manager_v2 * tm = static_cast<Trace_Manager_v2 *> (trace_manager);

  if (num_iterations > tm->get_num_traces_available (key_num))
    {
      std::cout << "W - Only " << std::dec << tm->get_num_traces_available (key_num)
		<< " traces available for selected key!" << std::endl;

      this->num_iterations = tm->get_num_traces_available (key_num);
    }
  else if (num_iterations == 0)
    this->num_iterations = tm->get_num_traces_available (key_num);
  else
    this->num_iterations = num_iterations;

  return this->num_iterations;
}


/**
 * Set the number of the key to use for this execution of the attack.
 * Must be called before set_num_iterations and run_contest
 *
 * \arg \c key_num The number for the key to use (related to the index file)
 * \return \c true if the key is available
 */
bool Contest_v2::set_key_num (unsigned int key_num)
{
  Trace_Manager_v2 * tm = static_cast<Trace_Manager_v2 *> (trace_manager);

  if ((key_num > tm->get_num_keys_in_index ())
      || (! tm->is_key_available (key_num)))
    {
      std::cerr << "E - Key #" << std::dec << key_num << " is not available!" << std::endl;

      return false;
    }

  this->key_num = key_num;

  return tm->set_targeted_key (key_num);
}







/******************************************************************
 * Class Contest_v4_RSM
 ******************************************************************/

/** Numeric identifier of the first RSM implementation of DPA contest v4 */
const uint32_t Contest_v4_RSM::contest_v4_RSM_id = 2;


Contest_v4_RSM::Contest_v4_RSM (uint32_t contest_identifier,
				const std::string &trace_directory,
				const std::string &results_file,
				bool allow_overwrite_results_file)
: Contest (contest_identifier, trace_directory, 
	   results_file, allow_overwrite_results_file)
{
}

/**
 * Constructor
 *
 * \arg \c trace_directory Name of the directory contining the traces
 * \arg \c index_file Name of the index file
 * \arg \c results_file Name of the file that will contain the results
 * \arg \c allow_overwrite_results_file Do not stop if the results file already exists
 * \arg \c fork_mode Communication mode with the attack program (\c true: Fork mode, \c false: FIFO mode)
 * \arg \c attack_program_name If in fork mode, name of the attack program, if in FIFO mode, prefix for the FIFO names
 * \arg \c provide_offset If \c true, the offset is provided to the attack program, else, it is replace with 0
 * \arg \c check_traces_availability For each trace in the index file, check whether the trace file exists. Default is \c true.
 */
Contest_v4_RSM::Contest_v4_RSM (const std::string &trace_directory,
				const std::string &index_file,
				const std::string &results_file,
				bool allow_overwrite_results_file,
				bool fork_mode,
				const std::string &attack_program_name,
				bool provide_offset,
				bool check_traces_availability)
: Contest (contest_v4_RSM_id, trace_directory, 
	   results_file, allow_overwrite_results_file)
{
  this->provide_offset = provide_offset;

  trace_manager = new Trace_Manager_v4_RSM (trace_directory, index_file,
					    check_traces_availability);
  trace_manager->init ();

  results_manager = new Results_Manager_AES (contest_identifier, 256);

  trace_raw_length = trace_manager->get_trace_raw_length ();
  results_raw_length = results_manager->get_results_raw_length ();

#ifdef _WIN32
  // FIFO mode only for Windows
  attack_comm = new Attack_Comm_FIFO (trace_raw_length, results_raw_length,
					  attack_program_name);
#else /* _WIN32 */
  if (fork_mode)
    {
      // Fork mode
      attack_comm = new Attack_Comm_Fork (trace_raw_length, results_raw_length,
					  attack_program_name);
    }
  else
    {
      // FIFO mode
      attack_comm = new Attack_Comm_FIFO (trace_raw_length, results_raw_length,
					  attack_program_name);
    }
#endif /* _WIN32 */

  trace_buffer = new int8_t[trace_raw_length];
  results_buffer = new int8_t[results_raw_length];
}


/**
 * Run the contest
 *
 * \return \c false if an error happened during the execution
 */
bool Contest_v4_RSM::run_contest ()
{
  bool res;

  Trace_Manager_v4_RSM * tm = static_cast<Trace_Manager_v4_RSM *> (trace_manager);

  const Trace_v4_RSM * trace;

  Timer timer;
  uint32_t duration_ms;
  float duration_s;

  assert (key_num >= 0);
  assert (num_iterations > 0);


  // Prepare the results file
  std::cout << "I - Preparing the results file..." << std::endl;
  res = results_manager->prepare_results_file (results_file, allow_overwrite_results_file);
  if (! res)
    return false;

  // Set up the number of traces
  results_manager->set_number_of_traces (num_iterations);

  // Prepare the communication with the program
  res = attack_comm->prepare_comm ();
  if (! res)
    return false;

  // Launch the attack!
  attack_comm->launch_attack_and_wait (num_iterations);
  if (! res)
    return false;

  // Main loop
  for (int trace_num = 0; trace_num < num_iterations; trace_num++)
    {
      // Retrieve the next trace
      std::cout << "I - Trace #" << std::dec << std::setw (6) << std::setfill ('0')
		<< trace_num << ": Reading trace" << std::flush;

      trace = tm->get_next_trace ();

      if (trace == NULL)
	{
	  std::cerr << std::endl << "E - Oops, no more trace to read?!" << std::endl;
	  break;
	}

      std::cout << " / Sending trace" << std::flush;

      // Retrieve the raw representation of the trace
      trace->get_raw (trace_buffer, provide_offset);

      // Start the timer
      timer.start ();

      // Send the trace to the attack
      attack_comm->send_trace (trace_buffer,
			       (trace_num == num_iterations - 1));

      std::cout << " / Waiting for results" << std::flush;

      // Receive the results from the attack
      attack_comm->receive_results (results_buffer,
				    (trace_num == num_iterations - 1));

      // Stop the timer
      duration_ms = timer.stop ();

      std::cout << " / Saving results" << std::flush;

      // Save the results
      results_manager->save_raw_results (results_buffer, *trace, duration_ms);

      duration_s = (float) duration_ms / 1000.;

      std::cout << " / Done ("
		<< std::fixed << std::setprecision (1) << duration_s
		<< " s) "
		<< results_manager->get_progression_string ()
		<< std::endl;
    }


  // Close the results file
  std::cout << "I - Closing the results file..." << std::endl;
  results_manager->close_results_file ();

  // Close the comm with the attack program
  attack_comm->close_comm ();

  return true;
}


/**
 * Set the number of traces to use for the execution of the attack.
 * Must be called after set_key_num and before run_contest.
 *
 * \arg \c num_iterations Number of traces to use (0 to use all the available traces)
 * \return Number of traces that will really be used (num_iterations except if there are less traces available)
 */
int Contest_v4_RSM::set_num_iterations (unsigned int num_iterations)
{
  assert (key_num >= 0);

  Trace_Manager_v4_RSM * tm = static_cast<Trace_Manager_v4_RSM *> (trace_manager);

  if (num_iterations > tm->get_num_traces_available (key_num))
    {
      std::cout << "W - Only " << std::dec << tm->get_num_traces_available (key_num)
		<< " traces available for selected key!" << std::endl;

      this->num_iterations = tm->get_num_traces_available (key_num);
    }
  if (num_iterations == 0)
    this->num_iterations = tm->get_num_traces_available (key_num);
  else
    this->num_iterations = num_iterations;

  return this->num_iterations;
}


/**
 * Set the number of the key to use for this execution of the attack.
 * Must be called before set_num_iterations and run_contest
 *
 * \arg \c key_num The number for the key to use (related to the index file)
 * \return \c true if the key is available
 */
bool Contest_v4_RSM::set_key_num (unsigned int key_num)
{
  Trace_Manager_v4_RSM * tm = static_cast<Trace_Manager_v4_RSM *> (trace_manager);

  if ((key_num > tm->get_num_keys_in_index ())
      || (! tm->is_key_available (key_num)))
    {
      std::cerr << "E - Key #" << std::dec << key_num << " is not available!" << std::endl;

      return false;
    }

  this->key_num = key_num;

  return tm->set_targeted_key (key_num);
}





/******************************************************************
 * Class Contest_v4_2
 ******************************************************************/

/** Numeric identifier of the DPA contest v4.2 */
const uint32_t Contest_v4_2::contest_v4_2_id = 3;


/**
 * Constructor
 *
 * \arg \c trace_directory Name of the directory contining the traces
 * \arg \c index_file Name of the index file
 * \arg \c results_file Name of the file that will contain the results
 * \arg \c allow_overwrite_results_file Do not stop if the results file already exists
 * \arg \c fork_mode Communication mode with the attack program (\c true: Fork mode, \c false: FIFO mode)
 * \arg \c attack_program_name If in fork mode, name of the attack program, if in FIFO mode, prefix for the FIFO names
 * \arg \c provide_offset If \c true, the offset is provided to the attack program, else, it is replace with 0
 * \arg \c check_traces_availability For each trace in the index file, check whether the trace file exists. Default is \c true
 * \arg \c use_float_format Transfer samples as floats to the attack and take into account the vertical gain and vertical offset from the trace. Default is \c true
 */
Contest_v4_2::Contest_v4_2 (const std::string &trace_directory,
			    const std::string &index_file,
			    const std::string &results_file,
			    bool allow_overwrite_results_file,
			    bool fork_mode,
			    const std::string &attack_program_name,
			    bool provide_offset,
			    bool check_traces_availability,
			    bool use_float_format)
  : Contest_v4_RSM (contest_v4_RSM_id, trace_directory, 
		    results_file, allow_overwrite_results_file)
{
  this->provide_offset = provide_offset;

  trace_manager = new Trace_Manager_v4_2 (trace_directory, index_file,
					  check_traces_availability,
					  use_float_format);
  trace_manager->init ();

  results_manager = new Results_Manager_AES (contest_identifier, 128);

  trace_raw_length = trace_manager->get_trace_raw_length ();
  results_raw_length = results_manager->get_results_raw_length ();

#ifdef _WIN32
  // FIFO mode only for Windows
  attack_comm = new Attack_Comm_FIFO (trace_raw_length, results_raw_length,
					  attack_program_name);
#else /* _WIN32 */
  if (fork_mode)
    {
      // Fork mode
      attack_comm = new Attack_Comm_Fork (trace_raw_length, results_raw_length,
					  attack_program_name);
    }
  else
    {
      // FIFO mode
      attack_comm = new Attack_Comm_FIFO (trace_raw_length, results_raw_length,
					  attack_program_name);
    }
#endif /* _WIN32 */

  trace_buffer = new int8_t[trace_raw_length];
  results_buffer = new int8_t[results_raw_length];
}
