/*
 * Classes for management of the results of an attack
 *
 * Copyright (C) 2013 Guillaume Duc <guillaume.duc@telecom-paristech.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <stdint.h>

#include <vector>
#include <iostream>
#include <iomanip>
#include <cstdio>
#include <cstring>
#include <cerrno>
#include <cassert>

#include "results.h"
#include "aes.h"


/*******************************************************************
 * Class AES_Attack_Partial_Results
 *******************************************************************/


AES_Attack_Partial_Results::AES_Attack_Partial_Results ()
  : partial_results(16)
{
  for (unsigned int index = 0; index < partial_results.size (); index++)
    partial_results[index].resize (256);
}

/**
 * Return the length of the raw representation of the results
 *
 * \return The length (in bytes) of the raw representation of the results
 */
unsigned int AES_Attack_Partial_Results::get_raw_length () const
{
  return 1 + 16 * 256;
}

/**
 * Parse the raw results sent by the attack program
 *
 * \arg \c buffer Pointer to the raw results sent by the attack program
 */
void AES_Attack_Partial_Results::set_from_raw (const int8_t *buffer)
{
  targeted_subkey_num = (uint8_t) *buffer++;

  for (unsigned int byte = 0; byte < 16; byte++)
    for (unsigned int val = 0; val < 256; val++)
      partial_results[byte][val] = (uint8_t) *buffer++;
}


std::ostream & operator<< (std::ostream &os, const AES_Attack_Partial_Results &r)
{
  // Targeted sub key
  os << "Targeted subkey: " << std::dec << r.targeted_subkey_num << std::endl;

  // Array of values
  os << "Partial results:" << std::endl;
  os << "00 01 02 03 04 05 06 07 08 09 10 11 12 13 14 15" << std::endl;
  os << "-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --" << std::endl;

  for (unsigned int val = 0; val < 256; val++)
    {
      for (unsigned int byte = 0; byte < 16; byte++)
	{
	  os << std::hex << std::setw (2) << std::setfill ('0') << (unsigned int) r.partial_results[byte][val];
	  os << " ";
	}
      os << std::endl;
    }

  return os;
}



/*******************************************************************
 * Class Results_Manager_AES
 *******************************************************************/

/**
 * Version of the binary result file (1 was produced by the version of
 * the wrapper used during DPA contest v2)
 */
const uint32_t Results_Manager_AES::binary_format_version = 2;

const uint8_t Results_Manager_AES::aes_subkey_length = 16;


Results_Manager_AES::Results_Manager_AES (uint32_t contest_identifier, int aes)
  : correct_subkey (aes_subkey_length)
{
  results_file = NULL;
  headers_written = false;
  this->contest_identifier = contest_identifier;
  number_of_traces = 0;
  this->aes = aes;
}


/**
 * Return the length of the raw representation of the results
 *
 * \return The length (in bytes) of the raw representation of the results
 */
unsigned int Results_Manager_AES::get_results_raw_length () const
{
  return partial_results.get_raw_length ();
}


/**
 * Write the headers in the results file
 *
 * Format of the headers:
 * - uint32_t      version
 * - uint32_t      identifier of the contest / implementation
 * - uint32_t      number of partial results (number of traces used)
 * - uint8_t       index of the targetted subkey
 * - uint8_t       size (in bytes) of the correct key
 * - uint8_t[size] correct key
 * - uint8_t       size (in bytes) of the correct subkey
 * - uint8_t[size] correct subkey
 *
 * \arg \c trace A reference to the first trace used
 * \return \c true if the headers have been properly written
 */
bool Results_Manager_AES::write_headers (const Trace &trace)
{
  assert (results_file != NULL);
  assert (number_of_traces != 0);

  assert (sizeof (binary_format_version) == sizeof (uint32_t));
  assert (sizeof (contest_identifier) == sizeof (uint32_t));
  assert (sizeof (number_of_traces) == sizeof (uint32_t));
  assert (sizeof (aes_subkey_length) == sizeof (uint8_t));

  uint8_t key_len;
  std::vector<uint8_t> key;

  // Version
  fwrite (&binary_format_version, sizeof (binary_format_version), 1, results_file);

  // Contest ID
  fwrite (&contest_identifier, sizeof (contest_identifier), 1, results_file);

  // Number of traces
  fwrite (&number_of_traces, sizeof (number_of_traces), 1, results_file);

  // Index of targeted subkey
  fwrite (&(partial_results.targeted_subkey_num), sizeof (uint8_t), 1, results_file);

  // Correct key
  key_len = (uint8_t) trace.get_key_length ();
  key = trace.get_key ();
  assert (key_len == key.size ());

  fwrite (&key_len, sizeof (key_len), 1, results_file);
  for (unsigned int index = 0; index < key_len; index++)
    fwrite (&(key[index]), sizeof (uint8_t), 1, results_file);

  // Compute the correct subkey
  if (aes == 128)
    AES128_key_schedule (key, correct_subkey, partial_results.targeted_subkey_num);
  else
    AES256_key_schedule (key, correct_subkey, partial_results.targeted_subkey_num);

  // Correct subkey
  fwrite (&aes_subkey_length, sizeof (aes_subkey_length), 1, results_file);
  for (unsigned int index = 0; index < aes_subkey_length; index++)
    fwrite (&(correct_subkey[index]), sizeof (uint8_t), 1, results_file);

  headers_written = true;
  return true;
}

/**
 * Save a new partial result sent by the attack program
 *
 * Format in the file:
 * - uint8_t       size of the plaintext
 * - uint8_t[size] plaintext
 * - uint8_t       size of the ciphertext
 * - uint8_t[size] ciphertext
 * - uint32_t      duration of the processing (in milliseconds)
 * - size_subkey * uint8_t[256] result matrix
 *
 * \arg \c buffer Pointer to the raw results sent by the attack program
 * \arg \c trace  A reference to the trace processed
 * \arg \c dureation_ms The duration of the processing of the trace (in ms)
 * \return \c true if the results have been properly written
 */
bool Results_Manager_AES::save_raw_results (const int8_t *buffer,
					    const Trace &trace,
					    uint32_t duration_ms)
{
  std::vector<uint8_t> plaintext, ciphertext;
  uint8_t plaintext_len, ciphertext_len;
  bool res;

  assert (results_file != NULL);

  partial_results.set_from_raw (buffer);

  if (! headers_written)
    {
      res = write_headers (trace);
      if (!res)
	return false;
    }

  plaintext = trace.get_plaintext ();
  ciphertext = trace.get_ciphertext ();

  plaintext_len = (uint8_t) trace.get_plaintext_length ();
  ciphertext_len = (uint8_t) trace.get_ciphertext_length ();

  assert (plaintext_len == plaintext.size ());
  assert (ciphertext_len == ciphertext.size ());

  // Write plaintext
  fwrite (&plaintext_len, sizeof (plaintext_len), 1, results_file);
  for (unsigned int index = 0; index < plaintext.size (); index++)
    fwrite (&(plaintext[index]), sizeof (uint8_t), 1, results_file);

  // Write ciphertext
  fwrite (&ciphertext_len, sizeof (ciphertext_len), 1, results_file);
  for (unsigned int index = 0; index < ciphertext.size (); index++)
    fwrite (&(ciphertext[index]), sizeof (uint8_t), 1, results_file);

  // Write duration
  fwrite (&duration_ms, sizeof (duration_ms), 1, results_file);

  // Write results matrix
  for (unsigned int byte = 0; byte < partial_results.partial_results.size (); byte++)
    for (unsigned int val = 0; val < 256; val++)
      fwrite (&(partial_results.partial_results[byte][val]), sizeof (uint8_t), 1, results_file);

  return true;
}


/**
 * Open the file where the results will be written
 *
 * \arg \c filename Name of the file
 * \arg \c allow_overwrite \c False to abort if the destination file already exists
 * \return \c True if the file has been properly opened
 */
bool Results_Manager_AES::prepare_results_file (const std::string &filename,
						bool allow_overwrite)
{
  assert (results_file == NULL);

  results_filename = filename;

  if (! allow_overwrite)
    {
      // Check whether the results file exists
      results_file = fopen (filename.c_str (), "rb");
      if (results_file != NULL)
	{
	  fclose (results_file);
	  results_file = NULL;

	  std::cerr << "E - Results file (" << filename << ") already exists!" << std::endl;
	  std::cerr << "E - Abort (use --truncate to change this behaviour)" << std::endl;

	  return false;
	}
    }

  // Open the file for writing
  results_file = fopen (filename.c_str (), "wb");
  if (results_file == NULL)
    {
      std::cerr << "E - Error opening results file (" << filename << ") for writing!" << std::endl;
      std::cerr << "E - Reason: " << std::strerror (errno) << std::endl;

      return false;
    }

  return true;
}


/**
 * Close the results file
 */
void Results_Manager_AES::close_results_file ()
{
  if (results_file == NULL)
    return;

  fclose (results_file);
  results_file = NULL;
}


/**
 * Set the number of traces used during the attack (for the results file headers)
 *
 * \arg \c number_of_traces Number of traces used
 */
void Results_Manager_AES::set_number_of_traces (uint32_t number_of_traces)
{
  this->number_of_traces = number_of_traces;
}


/**
 * Return a string describing the progression of the attack
 */
const std::string Results_Manager_AES::get_progression_string () const
{
  std::string res = "[";

  for (unsigned int byte = 0; byte < correct_subkey.size (); byte++)
    {
      if (partial_results.partial_results[byte][0] == correct_subkey[byte])
	{
	  res.append("*");
	  continue;
	}

      if ((partial_results.partial_results[byte][1] == correct_subkey[byte]) ||
	  (partial_results.partial_results[byte][2] == correct_subkey[byte]) ||
	  (partial_results.partial_results[byte][3] == correct_subkey[byte]) ||
	  (partial_results.partial_results[byte][4] == correct_subkey[byte]) ||
	  (partial_results.partial_results[byte][5] == correct_subkey[byte]))
	{
	  res.append(".");
	  continue;
	}

      res.append(" ");
    }

  res.append ("]");

  return res;
}
