/*
 * Classes for management of the different format of traces
 *
 * Copyright (C) 2013 Guillaume Duc <guillaume.duc@telecom-paristech.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef TRACES_H
#define TRACES_H

#include <stdint.h>

#include <string>
#include <vector>
#include <list>
#include <iostream>


/*******************************************************************************
 * Trace
 *******************************************************************************/

class Trace
{
public:
  virtual unsigned int get_plaintext_length () const = 0;
  virtual unsigned int get_ciphertext_length () const = 0;
  virtual unsigned int get_key_length () const = 0;
  virtual unsigned int get_num_samples () const = 0;

  virtual unsigned int get_raw_length () const = 0;
  virtual void get_raw (int8_t *buffer) const = 0;

  virtual const std::vector<uint8_t> & get_plaintext() const = 0;
  virtual const std::vector<uint8_t> & get_ciphertext() const = 0;
  virtual const std::vector<uint8_t> & get_key() const = 0;
};


/*******************************************************************************
 * Traces v2
 *******************************************************************************/

class Trace_v2_metadata
{
public:
  std::vector<uint8_t> plaintext;
  std::vector<uint8_t> ciphertext;
  std::vector<uint8_t> key;
  std::string filename;
  std::string full_filename;

  Trace_v2_metadata ();
};

std::ostream & operator<< (std::ostream &os, const Trace_v2_metadata &m);


class Trace_v2 : public Trace
{
public:
  static const unsigned int plaintext_length;
  static const unsigned int ciphertext_length;
  static const unsigned int key_length;
  static const unsigned int num_samples;

  std::vector<uint8_t> plaintext;
  std::vector<uint8_t> ciphertext;
  std::vector<uint8_t> key;
  std::vector<int16_t> samples;

  Trace_v2 ();

  virtual unsigned int get_plaintext_length () const;
  virtual unsigned int get_ciphertext_length () const;
  virtual unsigned int get_key_length () const;
  virtual unsigned int get_num_samples () const;
  virtual unsigned int get_raw_length () const;
  virtual void get_raw (int8_t *buffer) const;

  virtual const std::vector<uint8_t> & get_plaintext() const;
  virtual const std::vector<uint8_t> & get_ciphertext() const;
  virtual const std::vector<uint8_t> & get_key() const;

  bool read_from_file (const Trace_v2_metadata &metadata);
  bool read_from_file (const std::string &filename);
};

std::ostream & operator<< (std::ostream &os, const Trace_v2 &t);


/*******************************************************************************
 * Traces v4 RSM (v4.1)
 *******************************************************************************/

class Trace_v4_RSM_metadata
{
public:
  std::vector<uint8_t> plaintext;
  std::vector<uint8_t> ciphertext;
  std::vector<uint8_t> key;
  uint8_t offset;
  std::string sub_directory;
  std::string filename;
  std::string full_filename;

  Trace_v4_RSM_metadata ();
  Trace_v4_RSM_metadata (unsigned int p_len, unsigned int c_len,
			 unsigned int k_len);

};

std::ostream & operator<< (std::ostream &os, const Trace_v4_RSM_metadata &m);


class Trace_v4_RSM : public Trace
{
  static const unsigned int start_offset;
  static const unsigned int wavedesc_length;
  static const unsigned int vertical_gain_offset;
  static const unsigned int vertical_offset_offset;

public:
  static const unsigned int plaintext_length;
  static const unsigned int ciphertext_length;
  static const unsigned int key_length;
  static const unsigned int num_samples;

  std::vector<uint8_t> plaintext;
  std::vector<uint8_t> ciphertext;
  std::vector<uint8_t> key;
  std::vector<int8_t> samples;
  uint8_t offset;
 
  float vertical_gain;
  float vertical_offset;

  Trace_v4_RSM ();
  Trace_v4_RSM (unsigned int p_len, unsigned int c_len,
		unsigned int k_len, unsigned int n_samples);

  virtual unsigned int get_plaintext_length () const;
  virtual unsigned int get_ciphertext_length () const;
  virtual unsigned int get_key_length () const;
  virtual unsigned int get_num_samples () const;
  virtual unsigned int get_raw_length () const;

  virtual void get_raw (int8_t *buffer) const;
  virtual void get_raw (int8_t *buffer, bool provide_offset) const;

  virtual const std::vector<uint8_t> & get_plaintext() const;
  virtual const std::vector<uint8_t> & get_ciphertext() const;
  virtual const std::vector<uint8_t> & get_key() const;

  virtual bool read_from_file (const Trace_v4_RSM_metadata &metadata);
  virtual bool read_from_file (const std::string &filename);
};

std::ostream & operator<< (std::ostream &os, const Trace_v4_RSM &t);

/*******************************************************************************
 * Traces v4.2
 *******************************************************************************/


class Trace_v4_2_metadata : public Trace_v4_RSM_metadata
{
public:
  std::vector<uint8_t> offsets;
  std::vector<uint8_t> shuffle0;
  std::vector<uint8_t> shuffle10;

  Trace_v4_2_metadata ();
};

std::ostream & operator<< (std::ostream &os, const Trace_v4_2_metadata &m);

/* Old format */
class Trace_v4_2 : public Trace_v4_RSM
{
public:
  static const unsigned int plaintext_length;
  static const unsigned int ciphertext_length;
  static const unsigned int key_length;
  static const unsigned int num_samples;

  static const unsigned int shuffle0_length;
  static const unsigned int shuffle10_length;
  static const unsigned int offsets_length;

  std::vector<uint8_t> offsets;
  std::vector<uint8_t> shuffle0;
  std::vector<uint8_t> shuffle10;

  Trace_v4_2 ();

  virtual unsigned int get_plaintext_length () const;
  virtual unsigned int get_ciphertext_length () const;
  virtual unsigned int get_key_length () const;
  virtual unsigned int get_num_samples () const;
  virtual unsigned int get_raw_length () const;

  virtual void get_raw (int8_t *buffer, bool provide_offset) const;
  virtual bool read_from_file (const Trace_v4_RSM_metadata &metadata);

  using Trace_v4_RSM::get_raw; // To have access to Trace_v4_RSM::get_raw (int8_t *)
  using Trace_v4_RSM::read_from_file; // To have access to Trace_v4_RSM::read_from_file (const std::string &)
};

std::ostream & operator<< (std::ostream &os, const Trace_v4_2 &t);


/* News format (uses vertical gain and offset) */
class Trace_v4_2_float : public Trace_v4_RSM
{
public:
  static const unsigned int plaintext_length;
  static const unsigned int ciphertext_length;
  static const unsigned int key_length;
  static const unsigned int num_samples;

  static const unsigned int shuffle0_length;
  static const unsigned int shuffle10_length;
  static const unsigned int offsets_length;

  std::vector<uint8_t> offsets;
  std::vector<uint8_t> shuffle0;
  std::vector<uint8_t> shuffle10;

  Trace_v4_2_float ();

  virtual unsigned int get_plaintext_length () const;
  virtual unsigned int get_ciphertext_length () const;
  virtual unsigned int get_key_length () const;
  virtual unsigned int get_num_samples () const;
  virtual unsigned int get_raw_length () const;

  virtual void get_raw (int8_t *buffer, bool provide_offset) const;
  virtual bool read_from_file (const Trace_v4_RSM_metadata &metadata);

  using Trace_v4_RSM::get_raw; // To have access to Trace_v4_RSM::get_raw (int8_t *)
  using Trace_v4_RSM::read_from_file; // To have access to Trace_v4_RSM::read_from_file (const std::string &)
};

std::ostream & operator<< (std::ostream &os, const Trace_v4_2_float &t);


/*******************************************************************************
 * Trace_Manager base class
 *******************************************************************************/

class Trace_Manager
{
public:
  virtual void init () = 0;
  virtual unsigned int get_trace_raw_length () const = 0;
};


/*******************************************************************************
 * Trace_Manager_v2
 *******************************************************************************/

class Trace_Manager_v2 : public Trace_Manager
{
  std::string base_directory;
  std::string index_filename;

  std::list<Trace_v2_metadata *> metadata_list;
  std::list<Trace_v2_metadata *>::const_iterator metadata_iterator;

  std::vector<std::vector<uint8_t> > keys_list_in_index;
  std::vector<std::vector<uint8_t> > keys_list_available;

  std::vector<uint8_t> targeted_key;
  unsigned int targeted_key_index;

  Trace_v2 *trace;

  unsigned int num_traces_in_index;
  unsigned int num_traces_available;
  unsigned int next_trace_num;

  bool check_traces_availability;

public:
  Trace_Manager_v2 (const std::string &base_dir,
		    const std::string &index_file,
		    bool check_traces_availability = true);

  virtual void init ();

  virtual const Trace_v2* get_next_trace ();

  virtual unsigned int get_num_traces_in_index () const;
  virtual unsigned int get_num_traces_available () const;
  virtual unsigned int get_num_traces_available (unsigned int key_num) const;
  virtual unsigned int get_num_keys_in_index () const;
  virtual unsigned int get_num_keys_available () const;

  virtual unsigned int get_trace_raw_length () const;

  virtual bool is_key_available (unsigned int key_num) const;

  virtual bool set_targeted_key (unsigned int key_num);
};


/*******************************************************************************
 * Trace_Manager_v4_RSM
 *******************************************************************************/

class Trace_Manager_v4_RSM : public Trace_Manager
{
protected:
  std::string base_directory;
  std::string index_filename;

  std::list<Trace_v4_RSM_metadata *> metadata_list;
  std::list<Trace_v4_RSM_metadata *>::const_iterator metadata_iterator;

  std::vector<std::vector<uint8_t> > keys_list_in_index;
  std::vector<std::vector<uint8_t> > keys_list_available;

  std::vector<uint8_t> targeted_key;
  unsigned int targeted_key_index;

  Trace_v4_RSM *trace;

  unsigned int num_traces_in_index;
  unsigned int num_traces_available;
  unsigned int next_trace_num;

  bool check_traces_availability;

public:
  Trace_Manager_v4_RSM (const std::string &base_dir,
			const std::string &index_file,
			bool check_traces_availability = true);

  virtual void init ();

  virtual const Trace_v4_RSM* get_next_trace ();

  virtual unsigned int get_num_traces_in_index () const;
  virtual unsigned int get_num_traces_available () const;
  virtual unsigned int get_num_traces_available (unsigned int key_num) const;
  virtual unsigned int get_num_keys_in_index () const;
  virtual unsigned int get_num_keys_available () const;

  virtual unsigned int get_trace_raw_length () const;

  virtual bool is_key_available (unsigned int key_num) const;

  virtual bool set_targeted_key (unsigned int key_num);
};


/*******************************************************************************
 * Trace_Manager_v4_2
 *******************************************************************************/

class Trace_Manager_v4_2 : public Trace_Manager_v4_RSM
{
  bool use_float_format;

public:
  Trace_Manager_v4_2 (const std::string &base_dir,
		      const std::string &index_file,
		      bool check_traces_availability = true,
		      bool use_float_format = true);

  virtual void init ();
};



#endif /* TRACES_H */
