/*
 * Small tool to convert traces to text
 *
 * Copyright (C) 2014 Guillaume Duc <guillaume.duc@telecom-paristech.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <stdint.h>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>
#include <string>
#include <cstdlib>

#include "traces.h"

// Contest id string
static const std::string V2_ID = "v2";
static const std::string V4_RSM_ID = "v4_RSM";
static const std::string V4_2_ID = "v4_2";

static std::string edition, trace_filename, result_filename;


void print_usage ()
{
  std::cerr << "Usage: traces2text edition trace result" << std::endl;
  std::cerr << "Where:" << std::endl;
  std::cerr << "edition is the identifier of the edition of the contest ("
	    << V2_ID << ", " << V4_RSM_ID << ", " << V4_2_ID <<  ")" << std::endl;
  std::cerr << "trace is the full filename of the trace file" << std::endl;
  std::cerr << "result is the filename where the textual representation of the trace will be stored" << std::endl;
}

void dump_trace_v2 (std::ofstream& os)
{
  Trace_v2 trace;
  bool res;

  res = trace.read_from_file (trace_filename);
  if (! res)
    {
      std::cerr << "Error reading the trace file!" << std::endl;
      return;
    }

  for (std::vector<int16_t>::const_iterator i = trace.samples.begin (); i != trace.samples.end (); i++)
    os << std::dec << (int) *i << std::endl;
}


void dump_trace_v4_rsm (std::ofstream& os)
{
  Trace_v4_RSM trace;
  bool res;

  res = trace.read_from_file (trace_filename);
  if (! res)
    {
      std::cerr << "Error reading the trace file!" << std::endl;
      return;
    }

  for (std::vector<int8_t>::const_iterator i = trace.samples.begin (); i != trace.samples.end (); i++)
    os << std::dec << (int) *i << std::endl;
}

void dump_trace_v4_2 (std::ofstream& os)
{
  Trace_v4_2 trace;
  bool res;

  res = trace.read_from_file (trace_filename);
  if (! res)
    {
      std::cerr << "Error reading the trace file!" << std::endl;
      return;
    }

  for (std::vector<int8_t>::const_iterator i = trace.samples.begin (); i != trace.samples.end (); i++)
    os << std::dec << (int) *i << std::endl;
}


int main (int argc, char **argv)
{
  if (argc < 4)
    {
      print_usage ();
      exit (EXIT_FAILURE);
    }

  edition = argv[1];
  trace_filename = argv[2];
  result_filename = argv[3];

  std::ofstream res_stream (result_filename.c_str (), std::ofstream::out | std::ofstream::trunc);

  if (edition == V2_ID)
    dump_trace_v2 (res_stream);
  else if (edition == V4_RSM_ID)
    dump_trace_v4_rsm (res_stream);
  else if (edition == V4_2_ID)
    dump_trace_v4_2 (res_stream);
  else
    {
      std::cerr << "Unknown edition of the contest!" << std::endl;
      print_usage ();
      exit (EXIT_FAILURE);
    }

  res_stream.close ();
}
