/*
 * Classes for management of the DPA contest v2 traces
 *
 * Copyright (C) 2009, 2013 Guillaume Duc <guillaume.duc@telecom-paristech.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <algorithm>
#include <cstdlib>
#include <cstdio>
#include <cassert>

#include "traces.h"
#include "utils.h"



/*******************************************************************
 * Class Trace_v2
 *******************************************************************/

const unsigned int Trace_v2::plaintext_length = 16;
const unsigned int Trace_v2::ciphertext_length = 16;
const unsigned int Trace_v2::key_length = 16;
const unsigned int Trace_v2::num_samples = 3253;


Trace_v2::Trace_v2 ()
: plaintext (plaintext_length, 0),
  ciphertext (ciphertext_length, 0),
  key (key_length, 0),
  samples (num_samples, 0)
{

}


/**
 * Return the length of the plaintext
 *
 * \return The length (in bytes) of the plaintext
 */
unsigned int Trace_v2::get_plaintext_length () const
{
  return plaintext_length;
}


/**
 * Return the length of the ciphertext
 *
 * \return The length (in bytes) of the ciphertext
 */
unsigned int Trace_v2::get_ciphertext_length () const
{
  return ciphertext_length;
}


/**
 * Return the length of the key
 *
 * \return The length (in bytes) of the key
 */
unsigned int Trace_v2::get_key_length () const
{
  return key_length;
}


/**
 * Return the number of samples in the trace
 *
 * \return The number of samples
 */
unsigned int Trace_v2::get_num_samples () const
{
  return num_samples;
}


/**
 * Return the size of the raw representation of the trace
 *
 * \return The length (in bytes) of the raw representation of the trace
 */
unsigned int Trace_v2::get_raw_length () const
{
  return plaintext_length
    + ciphertext_length
    + 2 * num_samples; // int16_t
}


/**
 * Create a raw representation (byte array) of the trace in order to
 * send it to the attack program
 *
 * \arg \c buffer A pointer to a buffer where the raw representation of the trace will
 * be written. The size of this buffer must be at least \c get_raw_length()
 */
void Trace_v2::get_raw (int8_t *buffer) const
{
  assert (plaintext.size () == plaintext_length);
  assert (ciphertext.size () == ciphertext_length);
  assert (samples.size () == num_samples);

  // Plaintext
  for (unsigned int i = 0; i < plaintext.size (); i++)
    *buffer++ = (int8_t) plaintext[i];

  // Ciphertext
  for (unsigned int i = 0; i < ciphertext.size (); i++)
    *buffer++ = (int8_t) ciphertext[i];

  // Samples (transfered in little endian: LSB first)
  for (unsigned int i = 0; i < samples.size (); i++)
    {
      *buffer++ = (int8_t) (samples[i] & 0xFF); // LSB
      *buffer++ = (int8_t) ((samples[i] & 0xFF00) >> 8); // MSB
    }
}


/**
 * Return the plaintext
 *
 * \return A vector of uint8_t containing the bytes of the plaintext
 */
const std::vector<uint8_t> & Trace_v2::get_plaintext() const
{
  return plaintext;
};


/**
 * Return the ciphertext
 *
 * \return A vector of uint8_t containing the bytes of the ciphertext
 */
const std::vector<uint8_t> & Trace_v2::get_ciphertext() const
{
  return ciphertext;
};


/**
 * Return the key
 *
 * \return A vector of uint8_t containing the bytes of the key
 */
const std::vector<uint8_t> & Trace_v2::get_key() const
{
  return key;
};


/**
 * Read a trace from the file
 *
 * \arg \c metadata A reference to the metadata of the trace (retrieve from the index file)
 * \return \c true if the trace has been proeprly read, \c false if something went wrong
 */
bool Trace_v2::read_from_file (const Trace_v2_metadata &metadata)
{
  assert (metadata.key.size () == key.size ());
  assert (metadata.plaintext.size () == plaintext.size ());
  assert (metadata.ciphertext.size () == ciphertext.size ());

  // First copy the metadata
  for (unsigned int i = 0; i < metadata.key.size (); i++)
    key[i] = metadata.key[i];
  for (unsigned int i = 0; i < metadata.plaintext.size (); i++)
    plaintext[i] = metadata.plaintext[i];
  for (unsigned int i = 0; i < metadata.ciphertext.size (); i++)
    ciphertext[i] = metadata.ciphertext[i];

  // Read the trace from the file
  return read_from_file (metadata.full_filename);
}


/**
 * Read a trace from the file (warning: this function does only the read from the file
 * and does not configure the plaintext, ciphertext, etc.)
 *
 * \arg \c filename The name of the file containing the trace
 * \return \c true if the trace has been proeprly read, \c false if something went wrong
 */
bool Trace_v2::read_from_file (const std::string &filename)
{
  std::string line;
  unsigned int ns = 0;

  // Open the trace file
  std::ifstream trace_stream (filename.c_str ());

  while (! trace_stream.eof ())
    {
      // Read a line from the trace file
      std::getline (trace_stream, line);

      // If start with # -> it is a comment
      if (line[0] == '#')
	continue;

      // Read the sample
      samples[ns] = (int16_t) atoi (line.c_str ());
      ns++;

      if (ns == num_samples)
	break;
    }

  // Close file
  trace_stream.close ();

  assert (ns == num_samples);

  return true;
}


std::ostream & operator<< (std::ostream &os, const Trace_v2 &t)
{
  // Key
  os << "Key: ";
  for (unsigned int i = 0; i < t.key.size (); i++)
    os << std::hex << std::setw (2) << std::setfill ('0') << (unsigned int) t.key[i];
  os << std::endl;

  // Plaintext
  os << "Plaintext: ";
  for (unsigned int i = 0; i < t.plaintext.size (); i++)
    os << std::hex << std::setw (2) << std::setfill ('0') << (unsigned int) t.plaintext[i];
  os << std::endl;

  // Ciphertext
  os << "Ciphertext: ";
  for (unsigned int i = 0; i < t.ciphertext.size (); i++)
    os << std::hex << std::setw (2) << std::setfill ('0') << (unsigned int) t.ciphertext[i];
  os << std::endl;

  // Samples
  os << "Samples: ";
  for (unsigned int i = 0; i < t.samples.size (); i++)
    os << std::dec << (int) t.samples[i] << std::endl;

  os << std::endl;

  return os;
}


/*******************************************************************
 * Class Trace_v2_metadata
 *******************************************************************/


Trace_v2_metadata::Trace_v2_metadata ()
: plaintext (Trace_v2::plaintext_length, 0),
  ciphertext (Trace_v2::ciphertext_length, 0),
  key (Trace_v2::key_length, 0)
{

}


std::ostream & operator<< (std::ostream &os, const Trace_v2_metadata &m)
{
  // Key
  for (unsigned int i = 0; i < m.key.size (); i++)
    os << std::hex << std::setw (2) << std::setfill('0') << (unsigned int) m.key[i];

  os << " ";

  // Plaintext
  for (unsigned int i = 0; i < m.plaintext.size (); i++)
    os << std::hex << std::setw (2) << std::setfill('0') << (unsigned int) m.plaintext[i];

  os << " ";

  // Ciphertext
  for (unsigned int i = 0; i < m.ciphertext.size (); i++)
    os << std::hex << std::setw (2) << std::setfill('0') << (unsigned int) m.ciphertext[i];

  os << " ";

  // Filename
  os << m.filename;

  return os;
}





/*******************************************************************
 * Class Trace_Manager_v2
 *******************************************************************/

Trace_Manager_v2::Trace_Manager_v2 (const std::string &base_dir,
				    const std::string &index_file,
				    bool check_traces_availability)
: base_directory (base_dir), index_filename (index_file),
	check_traces_availability (check_traces_availability)
{
  num_traces_in_index = 0;
  num_traces_available = 0;
  next_trace_num = 0;
  targeted_key_index = 0;

  trace = new Trace_v2 ();

  init ();
}


/**
 * Read and parse the index file containing the metadata
 */
void Trace_Manager_v2::init ()
{
  std::string line;
  std::string key, pt, ct, filename;

  std::vector<uint8_t> key_vect (Trace_v2::key_length, 0);

  std::cout << "I - Reading index file..." << std::endl;

  std::ifstream index_stream (index_filename.c_str ());

  while (! index_stream.eof ())
    {
      // Read a line from the index file
      std::getline (index_stream, line);

      if (line.size () <= 0)
	break;

      num_traces_in_index++;

      std::stringstream line_stream (line);

      // Retrieve the different fields (separated by ' ')
      std::getline (line_stream, key, ' ');
      std::getline (line_stream, pt, ' ');
      std::getline (line_stream, ct, ' ');
      std::getline (line_stream, filename, ' ');

      // Register the key if it is not already registered
      hex2uint8_vector (key, &key_vect);
      if (find (keys_list_in_index.begin (), keys_list_in_index.end (), key_vect) == keys_list_in_index.end ())
	keys_list_in_index.push_back (key_vect);

      // Check whether the file exists
      if (check_traces_availability)
	if (! file_exists (base_directory, filename))
	  continue;

      num_traces_available++;

      // Create the metadata
      Trace_v2_metadata * metadata = new Trace_v2_metadata ();

      hex2uint8_vector (key, &(metadata->key));
      hex2uint8_vector (pt, &(metadata->plaintext));
      hex2uint8_vector (ct, &(metadata->ciphertext));
      metadata->filename = filename;
      metadata->full_filename = base_directory + PATHSEP + filename;

      // Add the trace metadata to the list of available traces
      metadata_list.push_back (metadata);

      if (find (keys_list_available.begin (), keys_list_available.end (), key_vect) == keys_list_available.end ())
	keys_list_available.push_back (key_vect);
    }

  index_stream.close ();

  assert (num_traces_available == metadata_list.size ());

  // Sort the vector containing the list of keys (to keep the same behavior as in previous version)
  std::sort (keys_list_in_index.begin (), keys_list_in_index.end ());


  // Display some info
  std::cout << "D - Total number of traces in the index file = " << std::dec << get_num_traces_in_index () << std::endl;
  std::cout << "D - Total number of traces available = " << std::dec << get_num_traces_available () << std::endl;

  std::cout << "D - Total number of keys in the index file = " << std::dec << get_num_keys_in_index () << std::endl;
  std::cout << "D - Total number of keys available = " << std::dec << get_num_keys_available () << std::endl;

  for (unsigned int key = 0; key < get_num_keys_in_index (); key++)
    {
      std::cout << "D - Key #" << std::dec << std::setw (2) << std::setfill('0') << key << " (";

      for (unsigned int i = 0; i < keys_list_in_index[key].size (); i++)
	std::cout << std::hex << std::setw (2) << std::setfill('0') << (unsigned int) keys_list_in_index[key][i];

      if (is_key_available (key))
	std::cout << "): " << std::dec << get_num_traces_available (key) << " traces available" << std::endl;
      else
	std::cout << "): not available" << std::endl;
	  
    }

  metadata_iterator = metadata_list.begin ();
}


/**
 * Read the next trace
 *
 * \return A pointer to the next trace corresponding to the selected key, \c NULL if not available
 */
const Trace_v2* Trace_Manager_v2::get_next_trace ()
{
  assert (metadata_list.size () > 0);
  assert (targeted_key.size () > 0);

  bool res;

  // No remaining traces
  if (metadata_iterator == metadata_list.end ())
    return NULL;

  // Iterate until we find a trace with the correct targeted key
  while ((*metadata_iterator)->key != targeted_key)
    {
      metadata_iterator++;

      if (metadata_iterator == metadata_list.end ())
	return NULL;
    }

  assert (*metadata_iterator != NULL);

  // Read the trace
  res = trace->read_from_file (**metadata_iterator);

  next_trace_num++;
  metadata_iterator++;

  if (res)
    return trace;
  else
    return NULL;
}


/**
 * Return the number of traces in the index file
 *
 * \return The number of traces described in the index file
 */
unsigned int Trace_Manager_v2::get_num_traces_in_index () const
{
  return num_traces_in_index;
}

/**
 * Return the number of traces available (all the traces may not have been installed by the participant)
 *
 * \return The number of traces really available
 */
unsigned int Trace_Manager_v2::get_num_traces_available () const
{
  return num_traces_available;
}

/**
 * Return the number of traces in the index file for key #
 *
 * \arg \c key_num The index of the key (relative to the index file)
 * \return The number of traces available for the given key
 */
unsigned int Trace_Manager_v2::get_num_traces_available (unsigned int key_num) const
{
  unsigned int res = 0;
  std::vector<uint8_t> key;
  std::list<Trace_v2_metadata *>::const_iterator iter;

  if (key_num >= keys_list_in_index.size ())
    return 0;

  key = keys_list_in_index[key_num];

  for (iter = metadata_list.begin (); iter != metadata_list.end (); iter++)
    if ((*iter)->key == key)
	res++;

  return res;
}

/**
 * Return the number of different keys in the index file
 *
 * \return The number of keys in the index file
 */
unsigned int Trace_Manager_v2::get_num_keys_in_index () const
{
  return keys_list_in_index.size ();
}

/**
 * Return the number of different keys available (all the traces may not have been installed by the participant)
 *
 * \return The number of keys available
 */
unsigned int Trace_Manager_v2::get_num_keys_available () const
{
  return keys_list_available.size ();
}


/**
 * Return the size of the raw representation of the trace
 *
 * \return The length (in bytes) of the raw representation of the trace
 */

unsigned int Trace_Manager_v2::get_trace_raw_length () const
{
  return trace->get_raw_length ();
}


/**
 * Check whether the key is available (key_num refers to the index of the key in the
 * index file)
 *
 * \return \c true if the key \c key_num is available (i.e. if there are some traces available for this key)
 */
bool Trace_Manager_v2::is_key_available (unsigned int key_num) const
{
  std::vector<uint8_t> key;

  if (key_num >= keys_list_in_index.size ())
    return false;

  key = keys_list_in_index[key_num];

  if (find (keys_list_available.begin (), keys_list_available.end (), key) == keys_list_available.end ())
    return false;

  return true;
}

/**
 * Set the number of the key in the index to target
 *
 * \arg \c key_num The index of the key to target
 * \return \c true if there are some traces available for this key
 */
bool Trace_Manager_v2::set_targeted_key (unsigned int key_num)
{
  if (! is_key_available (key_num))
    return false;

  targeted_key_index = key_num;
  targeted_key = keys_list_in_index[key_num];

  return true;
}
