/*
 * Classes for management of the DPA contest v4.2 traces
 *
 * Copyright (C) 2014 Guillaume Duc <guillaume.duc@telecom-paristech.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <algorithm>
#include <cstdlib>
#include <cstdio>
#include <cassert>

#if defined HAVE_BZLIB_H || defined _WIN32
#include <bzlib.h>
#endif /* HAVE_BZLIB_H */

#include "traces.h"
#include "utils.h"


/*******************************************************************
 * Class Trace_v4_2
 *******************************************************************/

const unsigned int Trace_v4_2::plaintext_length = 16;
const unsigned int Trace_v4_2::ciphertext_length = 16;
const unsigned int Trace_v4_2::key_length = 16;
const unsigned int Trace_v4_2::num_samples = 1704402;

const unsigned int Trace_v4_2::shuffle0_length = 16;
const unsigned int Trace_v4_2::shuffle10_length = 16;
const unsigned int Trace_v4_2::offsets_length = 16;

Trace_v4_2::Trace_v4_2 ()
  : Trace_v4_RSM(plaintext_length, ciphertext_length, key_length, num_samples),
    offsets (offsets_length, 0),
    shuffle0 (shuffle0_length, 0),
    shuffle10 (shuffle10_length, 0)
{
}


/**
 * Return the length of the plaintext
 *
 * \return The length (in bytes) of the plaintext
 */
unsigned int Trace_v4_2::get_plaintext_length () const
{
  return plaintext_length;
}


/**
 * Return the length of the ciphertext
 *
 * \return The length (in bytes) of the ciphertext
 */
unsigned int Trace_v4_2::get_ciphertext_length () const
{
  return ciphertext_length;
}


/**
 * Return the length of the key
 *
 * \return The length (in bytes) of the key
 */
unsigned int Trace_v4_2::get_key_length () const
{
  return key_length;
}


/**
 * Return the number of samples in the trace
 *
 * \return The number of samples
 */
unsigned int Trace_v4_2::get_num_samples () const
{
  return num_samples;
}


/**
 * Return the size of the raw representation of the trace
 *
 * \return The length (in bytes) of the raw representation of the trace
 */
unsigned int Trace_v4_2::get_raw_length () const
{
  return plaintext_length
    + ciphertext_length
    + shuffle0_length
    + shuffle10_length
    + offsets_length
    + num_samples;
}


/**
 * Create a raw representation (byte array) of the trace in order to
 * send it to the attack program
 *
 * \arg \c buffer A pointer to a buffer where the raw representation of the trace will
 * be written. The size of this buffer must be at least \c get_raw_length()
 * \arg \c provide_offset If true, send the correct offset to the attack, else the offset
 * is forced to 0
 */
void Trace_v4_2::get_raw (int8_t *buffer, bool provide_offset) const
{
  assert (plaintext.size () == plaintext_length);
  assert (ciphertext.size () == ciphertext_length);
  assert (samples.size () == num_samples);
  assert (offsets.size () == offsets_length);
  assert (shuffle0.size () == shuffle0_length);
  assert (shuffle10.size () == shuffle10_length);

  // Plaintext
  for (unsigned int i = 0; i < plaintext.size (); i++)
    *buffer++ = (int8_t) plaintext[i];

  // Ciphertext
  for (unsigned int i = 0; i < ciphertext.size (); i++)
    *buffer++ = (int8_t) ciphertext[i];

  // Shuffle0
  for (unsigned int i = 0; i < shuffle0.size (); i++)
    if (provide_offset)
      *buffer++ = (int8_t) shuffle0[i];
    else
      *buffer++ = 0;

  // Shuffle10
  for (unsigned int i = 0; i < shuffle10.size (); i++)
    if (provide_offset)
      *buffer++ = (int8_t) shuffle10[i];
    else
      *buffer++ = 0;

  // Shuffle0
  for (unsigned int i = 0; i < offsets.size (); i++)
    if (provide_offset)
      *buffer++ = (int8_t) offsets[i];
    else
      *buffer++ = 0;

  // Samples
  for (unsigned int i = 0; i < samples.size (); i++)
    *buffer++ = (int8_t) samples[i];
}


/**
 * Read a trace from the file
 *
 * \arg \c metadata A reference to the metadata of the trace (retrieve from the index file)
 * \return \c true if the trace has been proeprly read, \c false if something went wrong
 */
bool Trace_v4_2::read_from_file (const Trace_v4_RSM_metadata &metadata)
{
  const Trace_v4_2_metadata &metadata2 = static_cast<const Trace_v4_2_metadata &> (metadata);

  assert (metadata2.key.size () == key.size ());
  assert (metadata2.plaintext.size () == plaintext.size ());
  assert (metadata2.ciphertext.size () == ciphertext.size ());
  assert (metadata2.offsets.size () == offsets.size ());
  assert (metadata2.shuffle0.size () == shuffle0.size ());
  assert (metadata2.shuffle10.size () == shuffle10.size ());

  // First copy the metadata
  for (unsigned int i = 0; i < metadata2.key.size (); i++)
    key[i] = metadata2.key[i];
  for (unsigned int i = 0; i < metadata2.plaintext.size (); i++)
    plaintext[i] = metadata2.plaintext[i];
  for (unsigned int i = 0; i < metadata2.ciphertext.size (); i++)
    ciphertext[i] = metadata2.ciphertext[i];
  for (unsigned int i = 0; i < metadata2.offsets.size (); i++)
    offsets[i] = metadata2.offsets[i];
  for (unsigned int i = 0; i < metadata2.shuffle0.size (); i++)
    shuffle0[i] = metadata2.shuffle0[i];
  for (unsigned int i = 0; i < metadata2.shuffle10.size (); i++)
    shuffle10[i] = metadata2.shuffle10[i];

  // Read the trace from the file
  return read_from_file (metadata2.full_filename);
}



std::ostream & operator<< (std::ostream &os, const Trace_v4_2 &t)
{
  // Key
  os << "Key: ";
  for (unsigned int i = 0; i < t.key.size (); i++)
    os << std::hex << std::setw (2) << std::setfill ('0') << (unsigned int) t.key[i];
  os << std::endl;

  // Plaintext
  os << "Plaintext: ";
  for (unsigned int i = 0; i < t.plaintext.size (); i++)
    os << std::hex << std::setw (2) << std::setfill ('0') << (unsigned int) t.plaintext[i];
  os << std::endl;

  // Ciphertext
  os << "Ciphertext: ";
  for (unsigned int i = 0; i < t.ciphertext.size (); i++)
    os << std::hex << std::setw (2) << std::setfill ('0') << (unsigned int) t.ciphertext[i];
  os << std::endl;

  // Offsets
  os << "Offsets: ";
  for (unsigned int i = 0; i < t.offsets.size (); i++)
    os << std::hex << std::setw (2) << std::setfill ('0') << (unsigned int) t.offsets[i];
  os << std::endl;

  // Shuffle0
  os << "Shuffle0: ";
  for (unsigned int i = 0; i < t.shuffle0.size (); i++)
    os << std::hex << std::setw (2) << std::setfill ('0') << (unsigned int) t.shuffle0[i];
  os << std::endl;

  // Shuffle10
  os << "Shuffle10: ";
  for (unsigned int i = 0; i < t.shuffle10.size (); i++)
    os << std::hex << std::setw (2) << std::setfill ('0') << (unsigned int) t.shuffle10[i];
  os << std::endl;

  // Samples
  os << "Samples: ";
  for (unsigned int i = 0; i < t.samples.size (); i++)
    os << std::dec << (int) t.samples[i] << std::endl;

  os << std::endl;

  return os;
}


/*******************************************************************
 * Class Trace_v4_2_metadata
 *******************************************************************/


Trace_v4_2_metadata::Trace_v4_2_metadata ()
  : Trace_v4_RSM_metadata(Trace_v4_2::plaintext_length, Trace_v4_2::ciphertext_length, Trace_v4_2::key_length),
    offsets (Trace_v4_2::offsets_length, 0),
    shuffle0 (Trace_v4_2::shuffle0_length, 0),
    shuffle10 (Trace_v4_2::shuffle10_length, 0)
{

}


std::ostream & operator<< (std::ostream &os, const Trace_v4_2_metadata &m)
{
  // Key
  for (unsigned int i = 0; i < m.key.size (); i++)
    os << std::hex << std::setw (2) << std::setfill('0') << (unsigned int) m.key[i];

  os << " ";

  // Plaintext
  for (unsigned int i = 0; i < m.plaintext.size (); i++)
    os << std::hex << std::setw (2) << std::setfill('0') << (unsigned int) m.plaintext[i];

  os << " ";

  // Ciphertext
  for (unsigned int i = 0; i < m.ciphertext.size (); i++)
    os << std::hex << std::setw (2) << std::setfill('0') << (unsigned int) m.ciphertext[i];

  os << " ";

  // Offsets
  os << "Offsets: ";
  for (unsigned int i = 0; i < m.offsets.size (); i++)
    os << std::hex << std::setw (2) << std::setfill ('0') << (unsigned int) m.offsets[i];

  os << " ";

  // Shuffle0
  os << "Shuffle0: ";
  for (unsigned int i = 0; i < m.shuffle0.size (); i++)
    os << std::hex << std::setw (2) << std::setfill ('0') << (unsigned int) m.shuffle0[i];

  os << " ";

  // Shuffle10
  os << "Shuffle10: ";
  for (unsigned int i = 0; i < m.shuffle10.size (); i++)
    os << std::hex << std::setw (2) << std::setfill ('0') << (unsigned int) m.shuffle10[i];

  os << " ";

  // Directory
  os << m.sub_directory;
  os << " ";

  // Filename
  os << m.filename;

  return os;
}


/*******************************************************************
 * Class Trace_Manager_v4_2
 *******************************************************************/

Trace_Manager_v4_2::Trace_Manager_v4_2 (const std::string &base_dir,
					const std::string &index_file,
					bool check_traces_availability,
					bool use_float_format)
  : Trace_Manager_v4_RSM (base_dir, index_file, check_traces_availability),
    use_float_format (use_float_format)
{
  //delete trace;

  if (use_float_format)
    trace = new Trace_v4_2_float ();
  else
    trace = new Trace_v4_2 ();
}


/**
 * Read and parse the index file containing the metadata
 */
void Trace_Manager_v4_2::init ()
{
  std::string line;
  std::string key, pt, ct, offsets, shuffle0, shuffle10, subdir, filename;

  std::vector<uint8_t> key_vect (trace->get_key_length (), 0);

  std::cout << "I - Reading index file (v4_2)..." << std::endl;

  std::ifstream index_stream (index_filename.c_str ());

  while (! index_stream.eof ())
    {
      // Read a line from the index file
      std::getline (index_stream, line);

      if (line.size () <= 0)
	break;

      num_traces_in_index++;

      std::stringstream line_stream (line);

      // Retrieve the different fields (separated by ' ')
      std::getline (line_stream, key, ' ');
      std::getline (line_stream, pt, ' ');
      std::getline (line_stream, ct, ' ');
      std::getline (line_stream, shuffle0, ' ');
      std::getline (line_stream, shuffle10, ' ');
      std::getline (line_stream, offsets, ' ');
      std::getline (line_stream, subdir, ' ');
      std::getline (line_stream, filename, ' ');

      // Register the key if it is not already registered
      hex2uint8_vector (key, &key_vect);
      if (find (keys_list_in_index.begin (), keys_list_in_index.end (), key_vect) == keys_list_in_index.end ())
	keys_list_in_index.push_back (key_vect);

      // Check whether the file exists
      if (check_traces_availability)
	if (! file_exists (base_directory, subdir, filename))
	  continue;

      num_traces_available++;

      // Create the metadata
      Trace_v4_2_metadata * metadata = new Trace_v4_2_metadata ();

      hex2uint8_vector (key, &(metadata->key));
      hex2uint8_vector (pt, &(metadata->plaintext));
      hex2uint8_vector (ct, &(metadata->ciphertext));
      hex2uint4_vector (shuffle0, &(metadata->shuffle0));
      hex2uint4_vector (shuffle10, &(metadata->shuffle10));
      hex2uint4_vector (offsets, &(metadata->offsets));

      metadata->sub_directory = subdir;
      metadata->filename = filename;
      metadata->full_filename = base_directory + PATHSEP + subdir + PATHSEP + filename;

      // Add the trace metadata to the list of available traces
      metadata_list.push_back (metadata);

      if (find (keys_list_available.begin (), keys_list_available.end (), key_vect) == keys_list_available.end ())
	keys_list_available.push_back (key_vect);
    }

  index_stream.close ();

  assert (num_traces_available == metadata_list.size ());

  // Display some info
  std::cout << "D - Total number of traces in the index file = " << std::dec << get_num_traces_in_index () << std::endl;
  std::cout << "D - Total number of traces available = " << std::dec << get_num_traces_available () << std::endl;

  std::cout << "D - Total number of keys in the index file = " << std::dec << get_num_keys_in_index () << std::endl;
  std::cout << "D - Total number of keys available = " << std::dec << get_num_keys_available () << std::endl;

  for (unsigned int key = 0; key < get_num_keys_in_index (); key++)
    {
      std::cout << "D - Key #" << std::dec << std::setw (2) << std::setfill('0') << key << " (";

      for (unsigned int i = 0; i < keys_list_in_index[key].size (); i++)
	std::cout << std::hex << std::setw (2) << std::setfill('0') << (unsigned int) keys_list_in_index[key][i];

      if (is_key_available (key))
	std::cout << "): " << std::dec << get_num_traces_available (key) << " traces available" << std::endl;
      else
	std::cout << "): not available" << std::endl;
	  
    }

  metadata_iterator = metadata_list.begin ();
}
