/*
 * Classes for management of the DPA contest v4.2 traces
 * (new using vertical gain and offset information)
 *
 * Copyright (C) 2014 Guillaume Duc <guillaume.duc@telecom-paristech.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <algorithm>
#include <cstdlib>
#include <cstdio>
#include <cassert>

#if defined HAVE_BZLIB_H || defined _WIN32
#include <bzlib.h>
#endif /* HAVE_BZLIB_H */

#include "traces.h"
#include "utils.h"

#define SIZEOF_FLOAT 4

/*******************************************************************
 * Class Trace_v4_2_float
 *******************************************************************/

const unsigned int Trace_v4_2_float::plaintext_length = 16;
const unsigned int Trace_v4_2_float::ciphertext_length = 16;
const unsigned int Trace_v4_2_float::key_length = 16;
const unsigned int Trace_v4_2_float::num_samples = 1704402;

const unsigned int Trace_v4_2_float::shuffle0_length = 16;
const unsigned int Trace_v4_2_float::shuffle10_length = 16;
const unsigned int Trace_v4_2_float::offsets_length = 16;

Trace_v4_2_float::Trace_v4_2_float ()
  : Trace_v4_RSM(plaintext_length, ciphertext_length, key_length, num_samples),
    offsets (offsets_length, 0),
    shuffle0 (shuffle0_length, 0),
    shuffle10 (shuffle10_length, 0)
{
}


/**
 * Return the length of the plaintext
 *
 * \return The length (in bytes) of the plaintext
 */
unsigned int Trace_v4_2_float::get_plaintext_length () const
{
  return plaintext_length;
}


/**
 * Return the length of the ciphertext
 *
 * \return The length (in bytes) of the ciphertext
 */
unsigned int Trace_v4_2_float::get_ciphertext_length () const
{
  return ciphertext_length;
}


/**
 * Return the length of the key
 *
 * \return The length (in bytes) of the key
 */
unsigned int Trace_v4_2_float::get_key_length () const
{
  return key_length;
}


/**
 * Return the number of samples in the trace
 *
 * \return The number of samples
 */
unsigned int Trace_v4_2_float::get_num_samples () const
{
  return num_samples;
}


/**
 * Return the size of the raw representation of the trace
 *
 * \return The length (in bytes) of the raw representation of the trace
 */
unsigned int Trace_v4_2_float::get_raw_length () const
{
  return plaintext_length
    + ciphertext_length
    + shuffle0_length
    + shuffle10_length
    + offsets_length
    + SIZEOF_FLOAT * num_samples;
}


/**
 * Create a raw representation (byte array) of the trace in order to
 * send it to the attack program
 *
 * \arg \c buffer A pointer to a buffer where the raw representation of the trace will
 * be written. The size of this buffer must be at least \c get_raw_length()
 * \arg \c provide_offset If true, send the correct offset to the attack, else the offset
 * is forced to 0
 */
void Trace_v4_2_float::get_raw (int8_t *buffer, bool provide_offset) const
{
  assert (sizeof (float) == SIZEOF_FLOAT);
  assert (plaintext.size () == plaintext_length);
  assert (ciphertext.size () == ciphertext_length);
  assert (samples.size () == num_samples);
  assert (offsets.size () == offsets_length);
  assert (shuffle0.size () == shuffle0_length);
  assert (shuffle10.size () == shuffle10_length);

  // Plaintext
  for (unsigned int i = 0; i < plaintext.size (); i++)
    *buffer++ = (int8_t) plaintext[i];

  // Ciphertext
  for (unsigned int i = 0; i < ciphertext.size (); i++)
    *buffer++ = (int8_t) ciphertext[i];

  // Shuffle0
  for (unsigned int i = 0; i < shuffle0.size (); i++)
    if (provide_offset)
      *buffer++ = (int8_t) shuffle0[i];
    else
      *buffer++ = 0;

  // Shuffle10
  for (unsigned int i = 0; i < shuffle10.size (); i++)
    if (provide_offset)
      *buffer++ = (int8_t) shuffle10[i];
    else
      *buffer++ = 0;

  // Shuffle0
  for (unsigned int i = 0; i < offsets.size (); i++)
    if (provide_offset)
      *buffer++ = (int8_t) offsets[i];
    else
      *buffer++ = 0;

  // Samples
  for (unsigned int i = 0; i < samples.size (); i++)
    {
      float sample_f = (samples[i] * vertical_gain) - vertical_offset;
      int8_t *buf = (int8_t *) &sample_f;

      // TODO : endianness problem
      *buffer++ = buf[0];
      *buffer++ = buf[1];
      *buffer++ = buf[2];
      *buffer++ = buf[3];
    }
}


/**
 * Read a trace from the file
 *
 * \arg \c metadata A reference to the metadata of the trace (retrieve from the index file)
 * \return \c true if the trace has been proeprly read, \c false if something went wrong
 */
bool Trace_v4_2_float::read_from_file (const Trace_v4_RSM_metadata &metadata)
{
  const Trace_v4_2_metadata &metadata2 = static_cast<const Trace_v4_2_metadata &> (metadata);

  assert (metadata2.key.size () == key.size ());
  assert (metadata2.plaintext.size () == plaintext.size ());
  assert (metadata2.ciphertext.size () == ciphertext.size ());
  assert (metadata2.offsets.size () == offsets.size ());
  assert (metadata2.shuffle0.size () == shuffle0.size ());
  assert (metadata2.shuffle10.size () == shuffle10.size ());

  // First copy the metadata
  for (unsigned int i = 0; i < metadata2.key.size (); i++)
    key[i] = metadata2.key[i];
  for (unsigned int i = 0; i < metadata2.plaintext.size (); i++)
    plaintext[i] = metadata2.plaintext[i];
  for (unsigned int i = 0; i < metadata2.ciphertext.size (); i++)
    ciphertext[i] = metadata2.ciphertext[i];
  for (unsigned int i = 0; i < metadata2.offsets.size (); i++)
    offsets[i] = metadata2.offsets[i];
  for (unsigned int i = 0; i < metadata2.shuffle0.size (); i++)
    shuffle0[i] = metadata2.shuffle0[i];
  for (unsigned int i = 0; i < metadata2.shuffle10.size (); i++)
    shuffle10[i] = metadata2.shuffle10[i];

  // Read the trace from the file
  return read_from_file (metadata2.full_filename);
}



std::ostream & operator<< (std::ostream &os, const Trace_v4_2_float &t)
{
  // Key
  os << "Key: ";
  for (unsigned int i = 0; i < t.key.size (); i++)
    os << std::hex << std::setw (2) << std::setfill ('0') << (unsigned int) t.key[i];
  os << std::endl;

  // Plaintext
  os << "Plaintext: ";
  for (unsigned int i = 0; i < t.plaintext.size (); i++)
    os << std::hex << std::setw (2) << std::setfill ('0') << (unsigned int) t.plaintext[i];
  os << std::endl;

  // Ciphertext
  os << "Ciphertext: ";
  for (unsigned int i = 0; i < t.ciphertext.size (); i++)
    os << std::hex << std::setw (2) << std::setfill ('0') << (unsigned int) t.ciphertext[i];
  os << std::endl;

  // Offsets
  os << "Offsets: ";
  for (unsigned int i = 0; i < t.offsets.size (); i++)
    os << std::hex << std::setw (2) << std::setfill ('0') << (unsigned int) t.offsets[i];
  os << std::endl;

  // Shuffle0
  os << "Shuffle0: ";
  for (unsigned int i = 0; i < t.shuffle0.size (); i++)
    os << std::hex << std::setw (2) << std::setfill ('0') << (unsigned int) t.shuffle0[i];
  os << std::endl;

  // Shuffle10
  os << "Shuffle10: ";
  for (unsigned int i = 0; i < t.shuffle10.size (); i++)
    os << std::hex << std::setw (2) << std::setfill ('0') << (unsigned int) t.shuffle10[i];
  os << std::endl;

  // Vertical gain
  os << "Vertical gain: " << std::scientific << std::setprecision(6) << t.vertical_gain << std::endl;

  // Vertical offset
  os << "Vertical offset: " << std::scientific << std::setprecision(6) << t.vertical_offset << std::endl;

  // Samples
  os << "Samples: ";
  for (unsigned int i = 0; i < t.samples.size (); i++)
    os << std::dec << (int) t.samples[i] << std::endl;

  os << std::endl;

  return os;
}
