/*
 * Misc functions used by the attack wrapper
 *
 * Copyright (C) 2010 Guillaume Duc <guillaume.duc@telecom-paristech.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <cstdlib>
#include <cstdio>
#include <limits>

#include "utils.h"



/**
 * Convert an hexadecimal string to a vector of uint8_t
 *
 * \arg \c hex A string containing hexadecimal digits
 * \arg \c vect A vector of uint8_t that will be filled
 */
void hex2uint8_vector (const std::string &hex, std::vector<uint8_t> * vect)
{
  uint8_t val;

  if ((hex.size () % 2) != 0)
    return;

  if (hex.size () == 0)
    return;

  if ((2 * vect->size ()) != hex.size ())
    return;

  for (unsigned int pos = 0; pos < vect->size (); pos++)
    {
      std::string substr (hex, pos * 2, 2);
      val = (uint8_t) strtoul (substr.c_str (), NULL, 16);

      vect->at (pos) = val;
    }
}

/**
 * Convert an hexadecimal string to a vector of uint8_t
 *
 * \arg \c hex A string containing hexadecimal digits
 * \arg \c vect A vector of uint8_t that will be filled
 */
void hex2uint4_vector (const std::string &hex, std::vector<uint8_t> * vect)
{
  uint8_t val;

  if (hex.size () == 0)
    return;

  if (vect->size () != hex.size ())
    return;

  for (unsigned int pos = 0; pos < vect->size (); pos++)
    {
      std::string substr (hex, pos, 1);
      val = (uint8_t) strtoul (substr.c_str (), NULL, 16);

      vect->at (pos) = val;
    }
}

/**
 * Check whether a file exists (filename)
 *
 * \arg \c filename Name of the file
 * \return \c true if the file exists
 */
bool file_exists (const std::string &filename)
{
  // TODO: use an optimized method if available (ex. stat)
  FILE *file;

  file = fopen (filename.c_str (), "r");

  if (file == NULL)
    return false;

  fclose (file);

  return true;
}


/**
 * Check whether a file exists (dir/filename)
 *
 * \arg \c dir Directory name
 * \arg \c filename File name
 * \return \c true if the file exists
 */
bool file_exists (const std::string &dir, const std::string &filename)
{
  std::string file = dir + PATHSEP + filename;

  return file_exists (file);
}


/**
 * Check whether a file exists (dir/sub_dir/filename)
 *
 * \arg \c dir Directory name
 * \arg \c sub_dir Subdirectory name
 * \arg \c filename File name
 * \return \c true if the file exists
 */
bool file_exists (const std::string &dir, const std::string &sub_dir,
		  const std::string &filename)
{
  std::string file = dir + PATHSEP + sub_dir + PATHSEP + filename;

  return file_exists (file);
}



/**
 * Start the timer
 */
void Timer::start ()
{
#ifdef _WIN32
  start_tick = GetTickCount ();
#else /* _WIN32 */
  gettimeofday (&start_time, NULL);
#endif /* _WIN32 */
}


/**
 * Stop the timer
 *
 * \return The number of milliseconds ellapsed since the last call to start ()
 */
uint32_t Timer::stop ()
{
#ifdef _WIN32
  DWORD stop_tick = GetTickCount ();

  if (stop_tick > start_tick)
    return stop_tick - start_tick;
  else
    return (std::numeric_limits<DWORD>::max () - start_tick) + stop_tick;

#else /* _WIN32 */

  uint64_t start_us, stop_us, diff_us;

  gettimeofday (&stop_time, NULL);

  start_us = start_time.tv_usec + (uint64_t) start_time.tv_sec * 1000000;
  stop_us = stop_time.tv_usec + (uint64_t) stop_time.tv_sec * 1000000;
  diff_us = stop_us - start_us;

  return diff_us / 1000;
#endif /* _WIN32 */
}
